﻿#include "BaseTrack.h"


BaseTrack::BaseTrack(unsigned int num_states, unsigned int num_obs):m_num_states(num_states), m_num_obs(num_obs)
{
    //需要在派生类中赋值几个算法矩阵
}

void BaseTrack::Init(const std::vector<float>& data)
{
    if (kf_ == nullptr)
        return;
    int size = data.size();
    if (size > m_num_obs)
        return;
    Eigen::VectorXf observation = Eigen::VectorXf::Zero(size);
    for (int i = 0; i < data.size(); i++)
        observation(i) = data[i];
    kf_->x_.head(size) << observation;
    hit_streak_++;
    m_update_count++;
}
void BaseTrack::Predict()
{
    if (kf_ == nullptr)
        return;
    kf_->Predict();

    // hit streak count will be reset
    if (coast_cycles_ > 0) {
        hit_streak_ = 0;
    }
    // accumulate coast cycle count
    coast_cycles_++;
}
void BaseTrack::Update(const std::vector<float>& data)
{
    if (kf_ == nullptr)
        return;
    // get measurement update, reset coast cycle count
    coast_cycles_ = 0;
    // accumulate hit streak count
    hit_streak_++;
    m_update_count++;

    // observation - center_x, center_y, area, ratio
    int size = data.size();
    Eigen::VectorXf observation = Eigen::VectorXf::Zero(size);
    for (int i = 0; i < data.size(); i++)
        observation(i) = data[i];
    kf_->Update(observation);
}
void BaseTrack::UpdateDataCheck(const std::vector<float>& data, std::vector<float>& out)
{

}
void BaseTrack::UpdateHit()
{
    if (kf_ == nullptr)
        return;
    // get measurement update, reset coast cycle count
    coast_cycles_ = 0;
    // accumulate hit streak count
    hit_streak_++;
    m_update_count++;
}

int BaseTrack::GetStateData(std::vector<float>& data)
{
    if (kf_ == nullptr)
        return -1;
    data.clear();
    for (int i = 0; i < m_num_states; i++)
        data.push_back(kf_->x_[i]);
    return 0;
}
int BaseTrack::GetPredictData(std::vector<float>& data)
{
    if (kf_ == nullptr || m_num_obs > m_num_states)
        return -1;
    data.clear();
    for (int i = 0; i < m_num_obs; i++)
        data.push_back(kf_->x_predict_[i]);
    return 0;
}

float BaseTrack::GetNIS() const
{
    if (kf_ == nullptr)
        return 0;
    return kf_->NIS_;
}
float BaseTrack::GetProb() const
{
    return m_prob;
}

bool BaseTrack::IsLost()
{
    return coast_cycles_ > m_kMaxCoastCycles;
}
bool BaseTrack::IsValid()
{
    return m_update_count >= m_updateValidCount;
}
int BaseTrack::GetIouData(std::vector<float>& data, int& obj_type)
{
    if (kf_ == nullptr || m_num_obs > m_num_states)
        return -1;
    data.clear();
    for (int i = 0; i < m_num_obs; i++)
        data.push_back(kf_->x_[i]);
    return 0;
}

int BaseTrack::GetStatesNum()
{
    if (kf_ == nullptr)
        return 0;
    return kf_->num_states_;
}
int BaseTrack::GetObsNum()
{
    if (kf_ == nullptr)
        return 0;
    return kf_->num_obs_;
}

float* BaseTrack::GetStatesXPtr()
{
    if (kf_ == nullptr)
        return nullptr;
    return kf_->x_.data();
}
float* BaseTrack::GetPredictPtr()
{
    if (kf_ == nullptr)
        return nullptr;
    return kf_->P_.data();
}
float* BaseTrack::GetRPtr()
{
    if (kf_ == nullptr)
        return nullptr;
    return kf_->R_.data();
}
