#pragma once

#include "SafeQueue.hpp"
#include "track.h"
#include "transform_uv.h"
#include "yolov5.hpp"
#include <jfx_common_msgs/InferReses.h>
#include <jfx_common_msgs/InferResult.h>
#include <jfx_common_msgs/InferResults.h>
#include <ros/ros.h>
#include <time.h>

#include "vision_config.h"
#include "yaml-cpp/yaml.h"
#include <chrono>
using namespace std;
using namespace std::chrono;

namespace jfx_vision {

enum ObjType {
    OBJ_UnKnown = 0,
    OBJ_Person = 1,
    OBJ_Car = 2,
    OBJ_Bicycle = 3,
    OBJ_Motorbike = 4,
    OBJ_Cone = 5,
    OBJ_Bucket = 6,
    OBJ_Guard = 7,
    OBJ_Truck = 8,
    OBJ_Bus = 9
};

// carType
enum carType {
    OBJ_MyCar = 1,
    OBJ_Fence_truck = 2,
    OBJ_Large_sized_bus = 3,
    OBJ_Large_truck = 10,
    OBJ_Light_passenger_vehicle = 4,
    OBJ_Minibus = 5,
    OBJ_Minivan = 6,
    OBJ_MPV = 7,
    OBJ_Pickup_truck = 8,
    OBJ_Sedan = 9,
    OBJ_Small_sized_truck = 10,
    OBJ_SUV = 11,
    OBJ_Tanker = 12
};

std::unordered_map<std::string, ObjType> g_DetTypeMap = {
    { "person", ObjType::OBJ_Person },
    { "car", ObjType::OBJ_Car },
    { "bicycle", ObjType::OBJ_Bicycle },
    { "motorbike", ObjType::OBJ_Motorbike },
    { "cone", ObjType::OBJ_Cone },
    { "bucket", ObjType::OBJ_Bucket },
    { "guard", ObjType::OBJ_Guard },
    { "fence_truck", ObjType::OBJ_Truck },
    { "large_sized_bus", ObjType::OBJ_Bus },
    { "large_truck", ObjType::OBJ_Truck },
    { "light_passenger_vehicle", ObjType::OBJ_Car },
    { "minibus", ObjType::OBJ_Bus },
    { "minivan", ObjType::OBJ_Car },
    { "MPV", ObjType::OBJ_Car },
    { "pickup_truck", ObjType::OBJ_Car },
    { "sedan", ObjType::OBJ_Car },
    { "small_sized_truck", ObjType::OBJ_Truck },
    { "SUV", ObjType::OBJ_Car },
    { "tanker", ObjType::OBJ_Truck },
    { "xijing_truck", ObjType::OBJ_Car },
    { "others", ObjType::OBJ_Truck }
};

std::map<int, std::string> classmap = { { 0, "person" }, { 1, "car" }, { 2, "bicycle" }, { 3, "motorbike" },
    { 4, "cone" }, { 5, "bucket" }, { 6, "guard" }, { 7, "fence_truck" },
    { 8, "large_sized_bus" }, { 9, "large_truck" }, { 10, "light_passenger_vehicle" }, { 11, "minibus" },
    { 12, "minivan" }, { 13, "MPV" }, { 14, "others" }, { 15, "pickup_truck" },
    { 16, "sedan" }, { 17, "small_sized_truck" }, { 18, "SUV" }, { 19, "tanker" },
    { 20, "xijing_truck" } };

std::unordered_map<std::string, carType> g_GetSubTypeMap = {
    { "car", carType::OBJ_MyCar },
    { "fence_truck", carType::OBJ_Fence_truck },
    { "large_sized_bus", carType::OBJ_Large_sized_bus },
    { "large_truck", carType::OBJ_Large_truck },
    { "light_passenger_vehicle", carType::OBJ_Light_passenger_vehicle },
    { "minibus", carType::OBJ_Minibus },
    { "minivan", carType::OBJ_Minivan },
    { "MPV", carType::OBJ_MPV },
    { "pickup_truck", carType::OBJ_Pickup_truck },
    { "sedan", carType::OBJ_Sedan },
    { "small_sized_truck", carType::OBJ_Small_sized_truck },
    { "SUV", carType::OBJ_SUV },
    { "tanker", carType::OBJ_Tanker },
    { "xijing_truck", carType::OBJ_Tanker },
};

template <typename T>
class Detector {

public:
    Detector(VisionConfig& config)
    {
        _config = config;
    }
    ~Detector()
    {
    }

    int32_t Init(ros::NodeHandle& nh, string model_path, int32_t batch_size)
    {
        m_results_pub = nh.advertise<jfx_common_msgs::InferReses>("/InferReses", 100);
        m_transform_uv = make_shared<jfx::TransformUV>();
        _yoloV5_inferPtr = make_shared<YoloV5>();
        _yoloV5_inferPtr->Init(model_path, batch_size);

        for (int i = 0; i < batch_size; i++) {
            Results* results = new Results;
            results_map_.insert(std::make_pair(i, results));
        }
        _is_running = true;
        _processThreadPtr = make_shared<thread>(&Detector::Loop, this);

        if (_ifUseLocalTrack) {
            _objTracker.init(nh, _config);
        }
    }

    int32_t Push(T& frame)
    {
        _queue.push(frame);
    }

private:
    int32_t Process(T& frame)
    {
        if (_is_running) {
            auto st = high_resolution_clock::now();

            // roi process
            std::map<int, uint64_t> tm;
            std::vector<cv::Mat> inputMat;
            const int cam_size = _config.globalConf.camera_size;
            for (int i = 0; i < cam_size; i++) {
                tm[i] = frame[i].ts;
                cv::Mat& _mat = frame[i].mat;
                cv::Rect roiRect;
                roiRect.x = _config.cameraParamConfigs_map[i].roiRect.x;
                roiRect.y = _config.cameraParamConfigs_map[i].roiRect.y;
                roiRect.width = _config.cameraParamConfigs_map[i].roiRect.width;
                roiRect.height = _config.cameraParamConfigs_map[i].roiRect.height;

                cv::Mat roiImg = _mat(roiRect).clone();
                inputMat.push_back(roiImg);
            }

            _yoloV5_inferPtr->preprocess(inputMat);
            _yoloV5_inferPtr->infer();
            _yoloV5_inferPtr->post();
            auto results = _yoloV5_inferPtr->GetResults();

            // roi process

            for (int i = 0; i < cam_size; i++) {
                auto iter = results.find(i);
                if (iter == results.end()) {
                    continue;
                }
                cv::Rect roiRect;
                roiRect.x = _config.cameraParamConfigs_map[i].roiRect.x;
                roiRect.y = _config.cameraParamConfigs_map[i].roiRect.y;
                roiRect.width = _config.cameraParamConfigs_map[i].roiRect.width;
                roiRect.height = _config.cameraParamConfigs_map[i].roiRect.height;

                for (int j = 0; j < iter->second->results.size(); j++) {
                    auto& item = iter->second->results[j];
                    item.location[0] = item.location[0] + roiRect.x;
                    item.location[1] = item.location[1] + roiRect.y;
                    item.location[2] = item.location[2] + roiRect.x;
                    item.location[3] = item.location[3] + roiRect.y;
                    item.center_point[0] = item.center_point[0] + roiRect.x;
                    item.center_point[1] = item.center_point[1] + roiRect.y;
                    item.bottom_center_point[0] = item.bottom_center_point[0] + roiRect.x;
                    item.bottom_center_point[1] = item.bottom_center_point[1] + roiRect.y;
                }
            }

            // filter resultmap
            // clear resluts map
            for (int i = 0; i < results_map_.size(); i++) {
                results_map_[i]->results.clear();
            }

            for (int b = 0; b < cam_size; b++) {
                auto iter = results.find(b);
                if (iter == results.end()) {
                    continue;
                }
                for (int j = 0; j < iter->second->results.size(); j++) {
                    auto& item = iter->second->results[j];
                    int _width = _config.cameraParamConfigs_map[b].coordinate.intrinsics.width;
                    int _height = _config.cameraParamConfigs_map[b].coordinate.intrinsics.height;
                    if (check_edge(item, _width, _height)) {
                        continue;
                    }
                    MyResult result;
                    result.num_class = item.num_class;
                    result.location = item.location;
                    result.center_point = item.center_point;
                    result.bottom_center_point = item.bottom_center_point;
                    result.prob = item.prob;
                    results_map_.at(b)->results.push_back(result);
                }
                results_map_.at(b)->index = b;
            }

            if (_ifUseLocalTrack) {
                _objTracker.track(results_map_, tm);
            } else {
                SendRosMessage(results_map_);
            }

            auto millis = duration_cast<std::chrono::milliseconds>(high_resolution_clock::now() - st).count();
            ROS_DEBUG("Detector frame process time:%d ms", millis);
            logDetectorResults(results_map_, tm);
        }
    }

    bool check_edge(MyResult& item, int width, int height)
    {
        auto& top_x = item.location[0];
        auto& top_y = item.location[1];
        auto& bottom_x = item.location[2];
        auto& bottom_y = item.location[3];
        int edge_cut = _config.globalConf.edge_cut;
        if (top_x < edge_cut || top_y < edge_cut || (width - bottom_x) < edge_cut || (height - bottom_y) < edge_cut) {
            // JFLOG(JFINFO, "detect obj  near to edge, filter x1=%d y1=%d x2=%d y2=%d",
            //       top_x, top_y, bottom_x, bottom_y);
            return true;
        }
        return false;
    }

    void logDetectorResults(std::map<int, Results*>& result, std::map<int, uint64_t> tm)
    {
        for (int i = 0; i < results_map_.size(); i++) {
            int car_num = 0, truck_num = 0, bus_num = 0;
            auto iter = results_map_.find(i);
            if (iter == results_map_.end()) {
                ROS_ERROR("ResultsMap index=%d don't have", i);
                continue;
            }
            for (int j = 0; j < iter->second->results.size(); j++) {
                auto& item = iter->second->results[j];
                string name = classmap[item.num_class]; //"todo"; // todo item.facility_name.c_str();
                int type = g_DetTypeMap[name];

                switch (type) {
                case 2:
                    car_num++;
                    break;
                case 8:
                    truck_num++;
                    break;
                case 9:
                    bus_num++;
                    break;
                default:
                    break;
                }
            }
            ROS_DEBUG("Camera[%d] detect: time is %ld, ResultsSize  is %d, road is %s, carnum = %d, trucknum = %d, busnum = %d",
                results_map_.at(i)->index, tm[i], iter->second->results.size(), _config.cameraParamConfigs_map[i].id.c_str(), car_num, truck_num, bus_num);
        }
    }

    int32_t Loop()
    {
        uint64_t procFrmIdx = 0;

        while (_is_running) {
            T framePtr;
            if (!_queue.timeout_pop(framePtr, 100)) {
                continue;
            }

            procFrmIdx ++;
            if (_queue.size() > 10) {
                if (procFrmIdx % 2 == 0) {
                    ROS_INFO("Detector queue size=%ld, drop frame", _queue.size());
                    continue;
                }
            }

            Process(framePtr);
        }
    }

    int32_t SendRosMessage(std::map<int, Results*>& result)
    {
        jfx_common_msgs::InferResult msg_result = {};
        jfx_common_msgs::InferResults msg_results = {};
        jfx_common_msgs::InferReses msg_reses = {};

        for (auto iter = result.begin(); iter != result.end(); iter++) {
            msg_results.index = iter->second->index;
            std::vector<double> f_c_x_y;
            std::vector<float> transform_M;
            std::vector<double> dist;
            std::vector<float> r;
            f_c_x_y.emplace_back(_config.cameraParamConfigs_map.at(iter->second->index).coordinate.intrinsics.fx);
            f_c_x_y.emplace_back(_config.cameraParamConfigs_map.at(iter->second->index).coordinate.intrinsics.cx);
            f_c_x_y.emplace_back(_config.cameraParamConfigs_map.at(iter->second->index).coordinate.intrinsics.fy);
            f_c_x_y.emplace_back(_config.cameraParamConfigs_map.at(iter->second->index).coordinate.intrinsics.cy);
            transform_M = _config.cameraParamConfigs_map.at(iter->second->index).coordinate.transform_M;
            dist = _config.cameraParamConfigs_map.at(iter->second->index).coordinate.intrinsics.dist;
            m_transform_uv->Init(f_c_x_y, transform_M, dist);
            for (int i = 0; i < iter->second->results.size(); i++) {
                msg_result.num_class = iter->second->results[i].num_class;
                msg_result.location = iter->second->results[i].location;
                msg_result.prob = iter->second->results[i].prob;
                m_transform_uv->Process(
                    static_cast<double>(iter->second->results[i].center_point[0]),
                    static_cast<double>(iter->second->results[i].center_point[1]),
                    r);
                for (auto const& _r : r) {
                    msg_result.center_point.emplace_back(static_cast<float>(_r));
                }
                r.clear();
                m_transform_uv->Process(
                    static_cast<double>(iter->second->results[i].bottom_center_point[0]),
                    static_cast<double>(iter->second->results[i].bottom_center_point[1]),
                    r);
                for (auto const& _r : r) {
                    msg_result.bottom_center_point.emplace_back(static_cast<float>(_r));
                }
                r.clear();
                msg_result.lpr = "";
                msg_result.car_color = "";
                msg_results.results.emplace_back(msg_result);
                msg_result = {};
            }
            msg_reses.reses.emplace_back(msg_results);
            msg_results = {};
            f_c_x_y.clear();
            transform_M.clear();
            dist.clear();
        }

        msg_reses.header.frame_id = to_string(CalculateFrameId());
        msg_reses.header.stamp = ros::Time::now();
        m_results_pub.publish(msg_reses);
        msg_reses = {};
    }

    uint64_t CalculateFrameId()
    {
        if (std::numeric_limits<uint64_t>::max() != m_frame_id) {
            m_frame_id++;
        } else {
            m_frame_id = 0;
        }
        return m_frame_id;
    }

    inline uint64_t mills_UTC()
    {
        struct timespec t;
        clock_gettime(CLOCK_REALTIME, &t);
        return t.tv_sec * 1000ULL + t.tv_nsec / 1000000ULL;
    }

private:
    int32_t _frame_ounter = 0;
    int32_t _batch_size = 0;
    bool _is_running = false;

    std::shared_ptr<YoloV5> _yoloV5_inferPtr;

    std::shared_ptr<thread> _processThreadPtr;

    bool _ifUseLocalTrack = true;
    ObjTrack _objTracker;

    SafeQueue<T> _queue;

    // uv <--> gauss
    std::shared_ptr<jfx::TransformUV> m_transform_uv;

    // ros
    uint64_t m_frame_id = 0;
    ros::Publisher m_results_pub;
    VisionConfig _config;
    std::map<int, Results*> results_map_;
};

}
