#pragma once

#include "ros_image_sync_def.h"

#include "detector.h"
#include "vision_config.h"
#include <chrono>
#include <limits.h>
#include <mutex>
#include <string.h>
#include <yaml-cpp/yaml.h>

using namespace std;
using namespace std::chrono;

namespace jfx_vision {

using namespace message_filters;
using namespace message_filters::sync_policies;

typedef struct _tagSTFrame {
    int camIdx = 0;
    uint64_t ts = 0;
    cv::Mat mat;
} STFrame;

template <typename M0>
class ImageAsyncSubscriber {
    typedef Subscriber<M0> ImageSubscriber;
    typedef boost::shared_ptr<ImageSubscriber> ImageSubscriberPtr;

public:
    ImageAsyncSubscriber(ros::NodeHandle& nh, int cam_size)
    {
        init(nh, cam_size);
    };
    ~ImageAsyncSubscriber() {};

public:
    int init(ros::NodeHandle& nh, int cam_size)
    {
        this->cam_size = cam_size;
        this->frame_cnt = 0;
        ROS_INFO("ImageAsyncSubscriber CONSTR: cam_size %d", cam_size);

        this->drop_rate = g_visionConfig->globalConf.drop_rate;

        ROS_INFO("ImageAsyncSubscriber Init YoloV5 model path:%s", g_visionConfig->globalConf.detect_model_path.c_str());
        _detector.reset(new CVDetector(*g_visionConfig));
        _detector->Init(nh, g_visionConfig->globalConf.detect_model_path, cam_size);

        char szTopic[256] = { 0 };
        std::map<int, CameraParamConfig>::iterator itr = g_visionConfig->cameraParamConfigs_map.begin();
        for (; itr != g_visionConfig->cameraParamConfigs_map.end(); ++itr) {
            if (strstr((char*)itr->second.url.c_str(), "usb:") != nullptr) {
                // usb camera
                sprintf(szTopic, "/camera_array/cam%d/image_raw", itr->second.idx);
            } else {
                // ip camera
                sprintf(szTopic, "/cameras/%s/cam%d/image_raw", g_visionConfig->globalConf.node_name.c_str(), itr->second.idx);
            }
            ImageSubscriberPtr image(new ImageSubscriber(nh, szTopic, 10));
            cam_image_subs.push_back(image);
            image->registerCallback(boost::bind(&ImageAsyncSubscriber::asyncCallback, this, _1));
        }

        return 0;
    }

    void asyncCallback(const boost::shared_ptr<M0 const>& msg_img)
    {
        static uint64_t lastPostTS = ros::Time::now().toNSec() / 1000000;
        static uint32_t lastSeq[16] = { 0 };

        mtx.lock();

        try {
            STFrame frame;
            frame.camIdx = atoi(msg_img->header.frame_id.c_str());
            frame.ts = msg_img->header.stamp.toNSec() / 1000000;
            frame.mat = cv_bridge::toCvShare(msg_img, "bgr8")->image.clone();
            imgMap[frame.camIdx] = frame;

            if (msg_img->header.seq != 0) {
                if ((msg_img->header.seq - lastSeq[frame.camIdx]) > 1) {
                    ROS_WARN("lost image, cameraIdx=%d, ts=%ld, last seq=%d, cur seq=%d", frame.camIdx, frame.ts, msg_img->header.seq, lastSeq[frame.camIdx]);
                }

                lastSeq[frame.camIdx] = msg_img->header.seq;
                if (msg_img->header.seq % 100 == 0) {
                    ROS_INFO("camera %d frame seq reach %d", frame.camIdx, msg_img->header.seq);
                }
            }

            uint64_t curTS = ros::Time::now().toNSec() / 1000000;
            if (this->cam_size <= imgMap.size() || (curTS - lastPostTS) >= 150) {
                std::vector<STFrame> imgs;

                std::map<uint32_t, STFrame>::iterator itr = imgMap.end();
                for (int i = 0; i < this->cam_size; i++) {
                    itr = imgMap.find(i);
                    if (itr == imgMap.end()) {
                        STFrame tmpFrame;
                        tmpFrame.camIdx = i;
                        tmpFrame.ts = frame.ts;
                        int w = g_visionConfig->cameraParamConfigs_map.at(i).coordinate.intrinsics.width;
                        int h = g_visionConfig->cameraParamConfigs_map.at(i).coordinate.intrinsics.height;
                        cv::Mat zeroMat = cv::Mat::zeros(h, w, CV_8UC3);
                        tmpFrame.mat = zeroMat.clone();
                        imgs.push_back(tmpFrame);
                        ROS_INFO("push empty image of camera %d", i);
                    } else {
                        imgs.push_back(itr->second);
                    }
                }

                _detector->Push(imgs);

                lastPostTS = curTS;
                imgMap.clear();
            }
        } catch (...) {
            ROS_INFO("asyncCallback catch exception");
        }

        mtx.unlock();
    }

private:
    int cam_size;
    std::vector<ImageSubscriberPtr> cam_image_subs;
    std::map<uint32_t, STFrame> imgMap;
    std::mutex mtx;

    ros::Publisher image_infer_pub;

    uint32_t frame_cnt;
    uint32_t drop_rate;

    time_point<high_resolution_clock> star_clock;
    typedef Detector<vector<STFrame>> CVDetector;
    boost::shared_ptr<CVDetector> _detector;
};

}
