#pragma once

#include "ros_image_sync_def.h"

#include <chrono>
#include <limits.h>

#include "vision_config.h"
#include "detector.h"
#include <yaml-cpp/yaml.h>

using namespace std;
using namespace std::chrono;

namespace jfx_vision {

using namespace message_filters;
using namespace message_filters::sync_policies;

typedef struct _tagSTFrame {
    uint64_t ts = 0;
    cv::Mat mat;
} STFrame;

template <typename M0>
class MultiCamImageSynchronizer {
    typedef Subscriber<M0> ImageSubscriber;
    typedef boost::shared_ptr<ImageSubscriber> ImageSubscriberPtr;
    typedef Synchronizer<ApproximateTime<M0, M0>> TwoCamerasSynchronizer;
    typedef boost::shared_ptr<TwoCamerasSynchronizer> TwoCamerasSynchronizerPtr;
    typedef Synchronizer<ApproximateTime<M0, M0, M0>> ThreeCamerasSynchronizer;
    typedef boost::shared_ptr<ThreeCamerasSynchronizer> ThreeCamerasSynchronizerPtr;
    typedef Synchronizer<ApproximateTime<M0, M0, M0, M0>> FourCamerasSynchronizer;
    typedef boost::shared_ptr<FourCamerasSynchronizer> FourCamerasSynchronizerPtr;

public:
    MultiCamImageSynchronizer(ros::NodeHandle& nh, int cam_size)
    {
        // ros::NodeHandle nh("~");
        // int cam_size;
        // nh.getParam("cam_size", cam_size);
        this->cam_size = cam_size;
        this->frame_cnt = 0;
        ROS_INFO("MultiCamImageSynchronizer CONSTR: cam_size %d", cam_size);

        this->drop_rate = g_visionConfig->globalConf.drop_rate;

        ROS_INFO("MultiCamImageSynchronizer Init YoloV5 model path:%s", g_visionConfig->globalConf.detect_model_path.c_str());
        _detector.reset(new CVDetector(*g_visionConfig));
        _detector->Init(nh, g_visionConfig->globalConf.detect_model_path, cam_size);

        ImageSubscriberPtr image_0(new ImageSubscriber(nh, "/cameras/cam0/image_raw", 10));
        ImageSubscriberPtr image_1(new ImageSubscriber(nh, "/cameras/cam1/image_raw", 10));
        ImageSubscriberPtr image_2(new ImageSubscriber(nh, "/cameras/cam2/image_raw", 10));
        ImageSubscriberPtr image_3(new ImageSubscriber(nh, "/cameras/cam3/image_raw", 10));

        star_clock = high_resolution_clock::now();
        switch (cam_size) {
        case 1:
            cam_image_subs.push_back(image_0);
            image_0->registerCallback(boost::bind(&MultiCamImageSynchronizer::syncCallback1, this, _1));
            break;
        case 2:
            cam_image_subs.push_back(image_0);
            cam_image_subs.push_back(image_1);

            images_sync_2.reset(new TwoCamerasSynchronizer(ApproximateTime<M0, M0>(2), *(image_0.get()), *(image_1.get())));
            images_sync_2->registerCallback(boost::bind(&MultiCamImageSynchronizer::syncCallback2, this, _1, _2));
            break;
        case 3:
            cam_image_subs.push_back(image_0);
            cam_image_subs.push_back(image_1);
            cam_image_subs.push_back(image_2);
            images_sync_3.reset(new ThreeCamerasSynchronizer(ApproximateTime<M0, M0, M0>(3), *(image_0.get()), *(image_1.get()), *(image_2.get())));
            images_sync_3->registerCallback(boost::bind(&MultiCamImageSynchronizer::syncCallback3, this, _1, _2, _3));
            break;
        case 4:
            cam_image_subs.push_back(image_0);
            cam_image_subs.push_back(image_1);
            cam_image_subs.push_back(image_2);
            cam_image_subs.push_back(image_3);
            images_sync_4.reset(new FourCamerasSynchronizer(ApproximateTime<M0, M0, M0, M0>(4), *(image_0.get()), *(image_1.get()), *(image_2.get()), *(image_3.get())));
            images_sync_4->registerCallback(boost::bind(&MultiCamImageSynchronizer::syncCallback4, this, _1, _2, _3, _4));
            break;
        }
    };
    ~MultiCamImageSynchronizer() {};

private:
    int cam_size;
    std::vector<ImageSubscriberPtr> cam_image_subs;

    ros::Publisher image_infer_pub;
    TwoCamerasSynchronizerPtr images_sync_2;
    ThreeCamerasSynchronizerPtr images_sync_3;
    FourCamerasSynchronizerPtr images_sync_4;

    uint32_t frame_cnt;
    uint32_t drop_rate;

    time_point<high_resolution_clock> star_clock;
    typedef Detector<vector<STFrame>> CVDetector;
    boost::shared_ptr<CVDetector> _detector;

    void syncCallback1(const boost::shared_ptr<M0 const>& msg_img)
    {
        ROS_INFO("syncCallback1");
    }
    void syncCallback2(const boost::shared_ptr<M0 const>& msg_img0, const boost::shared_ptr<M0 const>& msg_img1)
    {
        frame_cnt++;
        auto fram_time = high_resolution_clock::now();
        auto millis = duration_cast<std::chrono::milliseconds>(fram_time - star_clock).count();
        float fps = 1000.0 / millis * frame_cnt;

        if (frame_cnt == INT_MAX) {
            frame_cnt = 0;
            star_clock = high_resolution_clock::now();
        }

        // yolov5 infer
        std::vector<STFrame> imgs;

        STFrame frm0, frm1;
        frm0.ts = msg_img0->header.stamp.toNSec() / 1000000;
        frm1.ts = msg_img1->header.stamp.toNSec() / 1000000;
        frm0.mat = cv_bridge::toCvShare(msg_img0, "bgr8")->image;
        frm1.mat = cv_bridge::toCvShare(msg_img1, "bgr8")->image;
        imgs.push_back(frm0);
        imgs.push_back(frm1);

        _detector->Push(imgs);
    }
    void syncCallback3(const boost::shared_ptr<M0 const>& msg_img0, const boost::shared_ptr<M0 const>& msg_img1,
        const boost::shared_ptr<M0 const>& msg_img2)
    {
        ROS_INFO("syncCallback3");
    }
    void syncCallback4(const boost::shared_ptr<M0 const>& msg_img0, const boost::shared_ptr<M0 const>& msg_img1,
        const boost::shared_ptr<M0 const>& msg_img2, const boost::shared_ptr<M0 const>& msg_img3)
    {
        frame_cnt++;
        auto fram_time = high_resolution_clock::now();
        auto millis = duration_cast<std::chrono::milliseconds>(fram_time - star_clock).count();
        float fps = 1000.0 / millis * frame_cnt;

        // ROS_INFO("sync FPS:%.1f", fps);
        if (frame_cnt == INT_MAX) {
            frame_cnt = 0;
            star_clock = high_resolution_clock::now();
        }

        //if ((drop_rate != 0) && (frame_cnt % drop_rate != 1)) {
        //    return;
        //}

        // yolov5 infer
        std::vector<STFrame> imgs;

        STFrame frm0, frm1, frm2, frm3;
        frm0.ts = msg_img0->header.stamp.toNSec() / 1000000;
        frm1.ts = msg_img1->header.stamp.toNSec() / 1000000;
        frm2.ts = msg_img2->header.stamp.toNSec() / 1000000;
        frm3.ts = msg_img3->header.stamp.toNSec() / 1000000;
        frm0.mat = cv_bridge::toCvShare(msg_img0, "bgr8")->image;
        frm1.mat = cv_bridge::toCvShare(msg_img1, "bgr8")->image;
        frm2.mat = cv_bridge::toCvShare(msg_img2, "bgr8")->image;
        frm3.mat = cv_bridge::toCvShare(msg_img3, "bgr8")->image;
        imgs.push_back(frm0);
        imgs.push_back(frm1);
        imgs.push_back(frm2);
        imgs.push_back(frm3);

        _detector->Push(imgs);
    }
};

}
