#pragma once

#include "config.hpp"
#include "coordinate_convert.h"
#include "framesaver.h"
#include "tracker.h"
#include "transform_uv.h"
#include "vision_config.h"
#include <chrono>
#include <jfx_common_msgs/det_tracking.h>
#include <jfx_common_msgs/det_tracking_array.h>
#include <list>
#include <map>
#include <mutex>
#include <opencv2/opencv.hpp>
#include <ros/ros.h>
#include <stdlib.h>
#include <thread>
#include <vector>
#include "BaseTracker.h"
#include "Track2D.h"

using namespace jfx_vision;

enum OutObjType {
    OutOBJ_UnKnown = 0,
    OutOBJ_Person = 1,
    OutOBJ_Car = 2,
    OutOBJ_Bicycle = 3,
    OutOBJ_Motorbike = 4,
    OutOBJ_Cone = 5,
    OutOBJ_Bucket = 6,
    OutOBJ_Guard = 7,
    OutOBJ_Truck = 8,
    OutOBJ_Bus = 9
};

typedef enum _tagENTrackState {
    enTSNotTrack = 0,
    enTSTracking = 1,
    enTSTracked = 2
} ENTrackState;

typedef struct _tagSTTrackItem {
    int cameraIdx = -1;
    uint64_t ts = 0;
    uint64_t batckIndex = 0;
    ENTrackState state = enTSNotTrack;
    std::vector<cv::Rect> detections;
    std::vector<TrackObj> detections_info;
} STTrackItem;

typedef struct _tagSTTrackBatch {
    uint64_t batckIndex = 0;
    bool bTrackOver = false;
    std::vector<STTrackItem> frameVec;
} STTrackBatch;

typedef struct _tagSTTrackState {
    uint64_t lastOutputMS = 0;
    uint64_t inObjNum = 0;
    uint64_t outObjNum = 0;
} TrackState;

typedef struct _tagSTTrackResource {
    CameraParamConfig cameraConf;
    std::thread* pThread = nullptr;
    std::list<STTrackItem> toTrackItemList;
    std::mutex trackItemMutex;

    jfx::TransformUV transUV;
    CoordinateConvert coordinatCvt;
} STTrackResource;

class ObjTrack {
public:
    ObjTrack() {};
    ~ObjTrack() {};

public:
    int init(ros::NodeHandle& nh, VisionConfig& config)
    {
        mVisionConfig = config;
        mFrameSaver.mVisionConfig = config;
        mRosPublisher = nh.advertise<jfx_common_msgs::det_tracking_array>("/TrackingRes", 100);

        mObjClassId2Type = {
            { 0, OutObjType::OutOBJ_Person }, { 1, OutObjType::OutOBJ_Car }, { 2, OutObjType::OutOBJ_Bicycle }, { 3, OutObjType::OutOBJ_Motorbike },
            { 4, OutObjType::OutOBJ_Cone }, { 5, OutObjType::OutOBJ_Bucket }, { 6, OutObjType::OutOBJ_Guard }, { 7, OutObjType::OutOBJ_Truck },
            { 8, OutObjType::OutOBJ_Bus }, { 9, OutObjType::OutOBJ_Truck }, { 10, OutObjType::OutOBJ_Car }, { 11, OutObjType::OutOBJ_Bus },
            { 12, OutObjType::OutOBJ_Car }, { 13, OutObjType::OutOBJ_Car }, { 14, OutObjType::OutOBJ_Truck }, { 15, OutObjType::OutOBJ_Car },
            { 16, OutObjType::OutOBJ_Car }, { 17, OutObjType::OutOBJ_Truck }, { 18, OutObjType::OutOBJ_Car }, { 19, OutObjType::OutOBJ_Truck },
            { 20, OutObjType::OutOBJ_Car }
        };

        mObjType2Name = {
            { OutObjType::OutOBJ_UnKnown, "Unknown" },
            { OutObjType::OutOBJ_Person, "Person" },
            { OutObjType::OutOBJ_Car, "Car" },
            { OutObjType::OutOBJ_Bicycle, "Bicycle" },
            { OutObjType::OutOBJ_Motorbike, "Motorbike" },
            { OutObjType::OutOBJ_Cone, "Cone" },
            { OutObjType::OutOBJ_Bucket, "Bucket" },
            { OutObjType::OutOBJ_Guard, "Guard" },
            { OutObjType::OutOBJ_Truck, "Truck" },
            { OutObjType::OutOBJ_Bus, "Bus" }
        };

        std::map<int, CameraParamConfig>::iterator itr = mVisionConfig.cameraParamConfigs_map.begin();
        for (; itr != mVisionConfig.cameraParamConfigs_map.end(); ++itr) {
            mCameraidx2ID[itr->first] = itr->second.id;
            
            auto res = std::make_shared<STTrackResource>();
            res->cameraConf = itr->second;
            res->pThread = new std::thread(&ObjTrack::threadTrack, this, itr->second.idx, res);
            mTrackResMap[itr->second.idx] = res;
        }

        mPubThread = new std::thread(&ObjTrack::threadPublish, this);

        return 0;
    }

    // 检测完的结果送入
    int track(std::map<int, Results*>& result, std::map<int, uint64_t>& tm)
    {
        cv::Rect rect;

        mBatchIndex++;

        std::map<int, Results*>::iterator itr = result.begin();
        for (; itr != result.end(); ++itr) {
            STTrackItem trackItem;
            trackItem.cameraIdx = itr->first;
            trackItem.batckIndex = mBatchIndex;
            trackItem.ts = tm[trackItem.cameraIdx];

            for (int i = 0; i < itr->second->results.size(); i++) {
                auto& item = itr->second->results[i];

                if (isVehicle(item.num_class)) {
                    if (item.prob < 0.35) {
                        continue;
                    }
                } else {
                    if (item.prob < 0.2) {
                        continue;
                    }
                }

                rect.x = item.location[0];
                rect.y = item.location[1];
                rect.width = item.location[2] - item.location[0];
                rect.height = item.location[3] - item.location[1];
                trackItem.detections.emplace_back(rect);

                TrackObj trackObj;
                trackObj.rect = rect;
                trackObj.type = item.num_class;
                trackObj.prob = item.prob;
                trackItem.detections_info.push_back(trackObj);

                mTrackState.inObjNum++;
            }

            std::map<int, std::shared_ptr<STTrackResource>>::iterator itrTrackRes = mTrackResMap.find(trackItem.cameraIdx);
            if (itrTrackRes != mTrackResMap.end()) {
                itrTrackRes->second->trackItemMutex.lock();
                itrTrackRes->second->toTrackItemList.push_back(trackItem);
                itrTrackRes->second->trackItemMutex.unlock();
            }
        }

        // mFrameSaver.save(result, tm);

        return 0;
    }

    bool isVehicle(int num_class)
    {
        if (num_class == 0 || num_class == 2 || num_class == 3) {
            return false;
        }

        return true;
    }

    // 跟踪线程
    void threadTrack(int cameraIndex, std::shared_ptr<STTrackResource> trackResPtr)
    {
        int nRet = 0;
        STTrackItem trackItem;
        Tracker tracker;
        std::vector<uint64_t> add_detection;
        std::vector<uint64_t> lost_detection;
        uint64_t trackFrmIdx = 0;

        BaseTracker<Track2D> _tracker;
        _tracker.SetGPU(0);
        _tracker.SetIouThreshold(0.1);
        _tracker.SetMaxCoastCycles(3);
        _tracker.SetValidUpdateCount(2);
        float high_low_rate = 0.4;

        initTransUV(trackResPtr->transUV, cameraIndex);
        initCoordCVT(trackResPtr->coordinatCvt, cameraIndex);

        while (mThreadRun) {
            // 从跟踪队列获取数据（公共）
            if (trackResPtr->toTrackItemList.size() <= 0) {
                std::this_thread::sleep_for(std::chrono::milliseconds(10));
                continue;
            }

            trackResPtr->trackItemMutex.lock();
            trackItem = *(trackResPtr->toTrackItemList.begin());
            trackResPtr->toTrackItemList.pop_front();
            trackFrmIdx ++;
            if (trackResPtr->toTrackItemList.size() > 10) {
                if (trackFrmIdx % 2 == 0) {
                    trackResPtr->toTrackItemList.pop_front();
                    ROS_INFO("Tracker queue size=%ld, drop frame", trackResPtr->toTrackItemList.size());
                }
            }
            trackResPtr->trackItemMutex.unlock();

            add_detection.clear();
            lost_detection.clear();

            std::vector< std::vector<float> > inputH;
            std::vector< std::vector<float> > inputL;
            std::vector<int> indexH;
            std::vector<int> indexL;

            for (int i = 0; i < trackItem.detections_info.size(); i++)
            {
                TrackObj& item = trackItem.detections_info[i];
                std::vector<float> data;
                auto width = static_cast<float>(trackItem.detections[i].width);
                auto height = static_cast<float>(trackItem.detections[i].height);
                data.push_back(trackItem.detections[i].x + trackItem.detections[i].width/2);
                data.push_back(trackItem.detections[i].y + trackItem.detections[i].height/2);
                data.push_back(width);
                data.push_back(height);
                if (item.prob > high_low_rate)
                {
                    inputH.push_back(data);
                    indexH.push_back(i);
                }
                else
                {
                    inputL.push_back(data);
                    indexL.push_back(i);
                }
            }
            std::map<uint64_t, int> updateHId;
            std::map<uint64_t, int> updateLId;
            std::vector<uint64_t> lostId;
            _tracker.Run(inputH, inputL, 4, 8, updateHId, updateLId, lostId);

            std::map<uint64_t, std::shared_ptr<Track2D> >& trackers = _tracker.GetStates();

            jfx_common_msgs::det_tracking_array objArray;
            for (auto& iter : trackers)
            {
                jfx_common_msgs::det_tracking obj = {};
                TrackObj info = {};
                obj.frame = trackItem.batckIndex;
                obj.create_stamp = trackItem.ts;
                int is_need_send = 0;//是否需要发送
                if (updateHId.find(iter.first) != updateHId.end() || updateLId.find(iter.first) != updateLId.end())
                {
                    if (updateHId.find(iter.first) != updateHId.end())
                        info = trackItem.detections_info[indexH[updateHId[iter.first]]];   //从检测结果里得到初始高分obj
                    if (updateLId.find(iter.first) != updateLId.end())
                        info = trackItem.detections_info[indexL[updateLId[iter.first]]];   //从检测结果里得到初始低分obj
                }
                else
                {
                    info = iter.second->m_obj;
                }
                obj.obj_id = iter.first;
                std::vector<float> data;
                if (iter.second->IsValid() && iter.second->GetStateData(data) == 0)   //初始obj更新跟踪的推理结果
                {
                    is_need_send = 1;
                    int bcenter_x = data[0] + data[2] / 2;
                    int bcenter_y = data[1] + data[3];
                    vector<float> r;
                    double lat = 0, lon = 0;
                    int ret = trackResPtr->transUV.Process(static_cast<double>(bcenter_x), static_cast<double>(bcenter_y), r);
                    if (ret == 0) {
                        //trkItem->m_NowX = r[0];
                        //trkItem->m_NowY = r[1];
                        jfx::Array result84;
                        jfx::Array objPoint = { r[0], r[1] };
                        trackResPtr->coordinatCvt.Dev2WGS84(objPoint, result84, false);
                        lat = result84[0];
                        lon = result84[1];
                    }
                    obj.score = info.prob;
                    obj.h = 0;
                    obj.w = 0;
                    obj.l = 0;
                    obj.Long = lon;
                    obj.Lat = lat;
                    // temp use
                    std::map<int, std::string>::iterator itr0 = mCameraidx2ID.find(trackItem.cameraIdx);
                    if (itr0 != mCameraidx2ID.end()) {
                        obj.color_name = itr0->second;
                    }

                    std::map<int, OutObjType>::iterator itr = mObjClassId2Type.find(info.type);
                    if (itr != mObjClassId2Type.end()) {
                        obj.type = itr->second;
                        std::map<OutObjType, std::string>::iterator itr2 = mObjType2Name.find(itr->second);
                        if (itr2 != mObjType2Name.end()) {
                            obj.name = itr2->second;
                        }
                    }

                    fillObjSize(obj);

                    objArray.array.emplace_back(obj);
                }


            }

            mPubDataMutex.lock();
            mPubDataList.emplace_back(objArray);
            mPubDataMutex.unlock();


            // 跟踪
            //tracker.Run(
            //    trackItem.detections, trackItem.detections_info,
            //    [&add_detection, this](const uint32_t& id, const TrackImplPtr& trkItem) {
            //        add_detection.push_back(id);
            //    },
            //    [&lost_detection, this](const uint32_t& id, const TrackImplPtr& trkItem) {
            //        lost_detection.push_back(id);
            //    });

            //// 目标打包
            //packRosObj(tracker, trackResPtr->transUV, trackResPtr->coordinatCvt, &trackItem);
        }
    }

    // ROS消息发布线程
    void threadPublish()
    {
        jfx_common_msgs::det_tracking_array objArray;

        while (mThreadRun) {
            printState();

            if (mPubDataList.empty()) {
                std::this_thread::sleep_for(std::chrono::milliseconds(10));
                continue;
            }

            mPubDataMutex.lock();
            objArray = *(mPubDataList.begin());
            mPubDataList.pop_front();
            mPubDataMutex.unlock();

            mTrackState.outObjNum += objArray.array.size();
            mRosPublisher.publish(objArray);
        }
    }

    // 初始化UV坐标转换
    void initTransUV(jfx::TransformUV& uv, int cameraIndex)
    {
        std::vector<double> f_c_x_y;
        std::vector<float> transform_M;
        std::vector<double> dist;

        f_c_x_y.emplace_back(mVisionConfig.cameraParamConfigs_map.at(cameraIndex).coordinate.intrinsics.fx);
        f_c_x_y.emplace_back(mVisionConfig.cameraParamConfigs_map.at(cameraIndex).coordinate.intrinsics.cx);
        f_c_x_y.emplace_back(mVisionConfig.cameraParamConfigs_map.at(cameraIndex).coordinate.intrinsics.fy);
        f_c_x_y.emplace_back(mVisionConfig.cameraParamConfigs_map.at(cameraIndex).coordinate.intrinsics.cy);
        transform_M = mVisionConfig.cameraParamConfigs_map.at(cameraIndex).coordinate.transform_M;
        dist = mVisionConfig.cameraParamConfigs_map.at(cameraIndex).coordinate.intrinsics.dist;

        uv.Init(f_c_x_y, transform_M, dist);
    }

    // 初始化84坐标转换
    void initCoordCVT(CoordinateConvert& cvt, int cameraIndex)
    {
        auto coord = mVisionConfig.cameraParamConfigs_map.at(cameraIndex).coordinate;
        cvt.Init(coord.horn_r, coord.rx, coord.ry, coord.horn_v, coord.vx, coord.vy);
    }

    /// 转换跟踪结果成ROS消息，放到发送队列
    void packRosObj(Tracker& tracker, jfx::TransformUV& transUV, CoordinateConvert& cvt, STTrackItem* pTrackItem)
    {
        if (!pTrackItem) {
            return;
        }

        jfx_common_msgs::det_tracking_array objArray;

        vector<float> r;
        auto& tracks = tracker.GetTracks();
        for (auto& trk : tracks) {
            auto& trkItem = trk.second;
            const auto& bbox = trkItem->GetStateAsBbox();

            int bcenter_x = bbox.x + bbox.width / 2;
            int bcenter_y = bbox.y + bbox.height;
            int ret = transUV.Process(static_cast<double>(bcenter_x), static_cast<double>(bcenter_y), r);
            if (ret == 0) {
                trkItem->m_NowX = r[0];
                trkItem->m_NowY = r[1];
                jfx::Array result84;
                jfx::Array objPoint = { r[0], r[1] };
                cvt.Dev2WGS84(objPoint, result84, false);
                trkItem->m_Lat = result84[0];
                trkItem->m_Lon = result84[1];
            }

            jfx_common_msgs::det_tracking obj;
            obj.frame = pTrackItem->batckIndex;
            obj.create_stamp = pTrackItem->ts;
            obj.score = trkItem->m_Prob;
            obj.h = 0;
            obj.w = 0;
            obj.l = 0;
            obj.obj_id = trkItem->m_Id;
            obj.Long = trkItem->m_Lon;
            obj.Lat = trkItem->m_Lat;

            // temp use
            std::map<int, std::string>::iterator itr0 = mCameraidx2ID.find(pTrackItem->cameraIdx);
            if (itr0 != mCameraidx2ID.end()) {
                obj.color_name = itr0->second;
            }

            std::map<int, OutObjType>::iterator itr = mObjClassId2Type.find(trkItem->m_Type);
            if (itr != mObjClassId2Type.end()) {
                obj.type = itr->second;
                std::map<OutObjType, std::string>::iterator itr2 = mObjType2Name.find(itr->second);
                if (itr2 != mObjType2Name.end()) {
                    obj.name = itr2->second;
                }
            }

            fillObjSize(obj);

            objArray.array.emplace_back(obj);
        }

        mPubDataMutex.lock();
        mPubDataList.emplace_back(objArray);
        mPubDataMutex.unlock();
    }

    /// 设置目标尺寸(经验值)
    void fillObjSize(jfx_common_msgs::det_tracking& obj)
    {
        double diff = (rand() % 3) * 0.1;
        switch (obj.type) {
        case OutObjType::OutOBJ_Truck:
        case OutObjType::OutOBJ_Bus:
            obj.l = 9.60718145 - diff;
            obj.h = 2.72270761 - diff;
            obj.w = 4.05017717 - diff;
            break;
        case OutObjType::OutOBJ_Car:
            obj.l = 4.60718145 - diff;
            obj.h = 1.72270761 - diff;
            obj.w = 1.95017717 - diff;
            break;
        case OutObjType::OutOBJ_Bicycle:
        case OutObjType::OutOBJ_Motorbike:
            obj.l = 2.09973778 - diff;
            obj.h = 1.44403034 - diff;
            obj.w = 0.76279481 - diff;
            break;
        case OutObjType::OutOBJ_Person:
            obj.l = 0.5356437 - diff;
            obj.h = 1.75748069 - diff;
            obj.w = 0.5256437 - diff;
            break;
        default:
            obj.l = 4.60718145 - diff;
            obj.h = 1.72270761 - diff;
            obj.w = 1.95017717 - diff;
        }
    }

    /// 打印统计信息
    void printState()
    {
        uint64_t msCur = mills_UTC();
        if ((msCur - mTrackState.lastOutputMS) > 10000) {
            std::string msgListInfo;
            std::map<int, std::shared_ptr<STTrackResource>>::iterator itr = mTrackResMap.begin();
            for (itr = mTrackResMap.begin(); itr != mTrackResMap.end(); ++itr) {
                ROS_INFO("camera: %d, in list size: %ld", itr->second->cameraConf.idx, itr->second->toTrackItemList.size());
            }

            ROS_INFO("in: %.2f objs/s, out: %.2f objs/s",
                (double)mTrackState.inObjNum * 1000 / (msCur - mTrackState.lastOutputMS),
                (double)mTrackState.outObjNum * 1000 / (msCur - mTrackState.lastOutputMS));
            mTrackState.lastOutputMS = msCur;
            mTrackState.inObjNum = 0;
            mTrackState.outObjNum = 0;
        }
    }

public:
    uint64_t mBatchIndex = 0;
    std::map<int, std::shared_ptr<STTrackResource>> mTrackResMap;

    std::list<jfx_common_msgs::det_tracking_array> mPubDataList;
    std::mutex mPubDataMutex;
    std::thread* mPubThread = nullptr;

    bool mThreadRun = true;

    ros::Publisher mRosPublisher;

    std::map<OutObjType, std::string> mObjType2Name;
    std::map<int, OutObjType> mObjClassId2Type;
    std::map<int, std::string> mCameraidx2ID;

    VisionConfig mVisionConfig;
    FrameSerialize mFrameSaver;

    TrackState mTrackState;
};
