import numpy as np
from juefx_iou import bev_overlap
from juefx_kalman import kalman_update
from juefx_kalman_batch import kalman_update_batch

def assert_type(x):
    assert x.dtype == "float32"


def iou3d_gpu(boxes_a, boxes_b):

    """
    boxes_a: N * 7
    boxes_b: M * 7
    
    [x,y,z,l,w,h,thetha]
    """
    m, n = len(boxes_a), len(boxes_b)
    bo =  np.zeros((m, n)).astype(np.float32)

    # height overlap
    boxes_a_height_max = (boxes_a[:, 2] + boxes_a[:, 5] / 2).reshape(-1, 1)
    boxes_a_height_min = (boxes_a[:, 2] - boxes_a[:, 5] / 2).reshape(-1, 1)
    boxes_b_height_max = (boxes_b[:, 2] + boxes_b[:, 5] / 2).reshape(1, -1)
    boxes_b_height_min = (boxes_b[:, 2] - boxes_b[:, 5] / 2).reshape(1, -1)

    # bev overlap

    max_of_min = np.maximum(boxes_a_height_min, boxes_b_height_min)
    min_of_max = np.minimum(boxes_a_height_max, boxes_b_height_max)
    overlaps_h = np.maximum(min_of_max - max_of_min, 0)

    bev_overlap(m, boxes_a, n, boxes_b, bo)

    o3d = bo * overlaps_h # 3 * 3
    vol_a = (boxes_a[:, 3] * boxes_a[:, 4] * boxes_a[:, 5]).reshape(-1, 1)
    vol_b = (boxes_b[:, 3] * boxes_b[:, 4] * boxes_b[:, 5]).reshape(1, -1)
    res = o3d / np.maximum(1e-6, vol_a + vol_b - o3d)
    return res


def iou2d_gpu(boxes_a, boxes_b):

    """
    boxes_a: N * 7
    boxes_b: M * 7
    
    [x,y,z,l,w,h,thetha]
    """

    m, n = len(boxes_a), len(boxes_b)
    bo =  np.zeros((m, n)).astype(np.float32)
    bev_overlap(m, boxes_a, n, boxes_b, bo)
    area1 = (boxes_a[:, 3] * boxes_a[:, 4]).reshape(-1, 1)
    area2 = (boxes_b[:, 3] * boxes_b[:, 4]).reshape(1, -1)
    res = bo / np.maximum(1e-6, area1 + area2 - bo)
    return res


def kalman_update_gpu(Z, H, R, X, P):

    """
    ns = dim_x                          len of state
    no = dim_z                          len of observation

    Z: measurement                                                              (dim_z, 1)
    H: measurement function                                                     (dim_z, dim_x)
    E / R: measurement uncertainty / noise, named "R" in python, "E" in CUDA    (dim_z, dim_z)
    X: estimate                                                                 (dim_x, 1)
    P: uncertainty covariance                                                   (dim_x, dim_x)

    MAKE SURE ALL INPUTS ARE FLOAT32!
    """

    assert X.shape[0] == P.shape[0] == P.shape[1] == H.shape[1]     # ns = dim_x
    assert Z.shape[0] == R.shape[0] == R.shape[1] == H.shape[0]     # no = dim_z
    assert X.shape[1] == Z.shape[1] == 1

    assert_type(Z)
    assert_type(H)
    assert_type(R)
    assert_type(X)
    assert_type(P)

    ns = X.shape[0]
    no = Z.shape[0]

    kalman_update(Z, H, R, X, P, ns, no)


def kalman_update_batch_gpu(Zs, Xs, Ps, bs, ns, no):

    """

    ns = dim_x = 10                          len of state
    no = dim_z = 7                          len of observation

    Zs: measurement                                                              (bs, 7)
    Xs: estimate                                                                 (bs, 10)
    Ps: uncertainty covariance                                                   (bs, 100)


    H: measurement function                                                     (dim_z, dim_x)
    E / R: measurement uncertainty / noise, named "R" in python, "E" in CUDA    (dim_z, dim_z)

    MAKE SURE ALL INPUTS ARE FLOAT32!
    """

    assert_type(Zs)
    assert_type(Xs)
    assert_type(Ps)

    assert Zs.shape[0] == Ps.shape[0] == Xs.shape[0] == bs
    assert Zs.shape[1] == no
    assert Xs.shape[1] == ns
    assert Ps.shape[1] == ns * ns

    # can not direct use reshape or slicing, must as follows:
    HXs = np.zeros((bs, no))
    HXs = Xs[:, :7].copy()


    kalman_update_batch(Zs, Xs, Ps, HXs, bs, ns, no)




