#from __future__ import absolute_import, division
import numpy as np
from numpy import dot, zeros, eye, isscalar
from filterpy.common import reshape_z
from gpu_helpers import kalman_update_gpu, kalman_update_batch_gpu 

class KalmanFilter_GPU(object):
    def __init__(self, dim_x, dim_z, dim_u=0):
        if dim_x < 1:
            raise ValueError('dim_x must be 1 or greater')
        if dim_z < 1:
            raise ValueError('dim_z must be 1 or greater')
        if dim_u < 0:
            raise ValueError('dim_u must be 0 or greater')

        self.dt = np.float32

        self.dim_x = dim_x
        self.dim_z = dim_z
        self.dim_u = dim_u

        self.x = zeros((dim_x, 1)).astype(self.dt)        # state
        self.P = eye(dim_x).astype(self.dt)                # uncertainty covariance
        self.Q = eye(dim_x).astype(self.dt)                # process uncertainty
        self.B = None                     # control transition matrix
        self.F = eye(dim_x).astype(self.dt)                # state transition matrix
        self.F_blindspot = eye(dim_x).astype(self.dt)
        self.H = zeros((dim_z, dim_x)).astype(self.dt)     # measurement function
        self.R = eye(dim_z).astype(self.dt)                # measurement uncertainty
        self._alpha_sq = 1.               # fading memory control
        self.M = np.zeros((dim_x, dim_z)).astype(self.dt)  # process-measurement cross correlation
        self.z = np.array([[None]*self.dim_z]).T.astype(self.dt) 

        # gain and residual are computed during the innovation step. We
        # save them so that in case you want to inspect them for various
        # purposes
        self.K = np.zeros((dim_x, dim_z)).astype(self.dt)  # kalman gain
        self.y = zeros((dim_z, 1)).astype(self.dt) 
        self.S = np.zeros((dim_z, dim_z)).astype(self.dt)  # system uncertainty
        self.SI = np.zeros((dim_z, dim_z)).astype(self.dt)  # inverse system uncertainty


    def predict(self, blindSpot=None, u=None, B=None, F=None, Q=None):
        """
        Predict next state (prior) using the Kalman filter state propagation
        equations.

        Parameters
        ----------

        u : np.array, default 0
            Optional control vector.

        B : np.array(dim_x, dim_u), or None
            Optional control transition matrix; a value of None
            will cause the filter to use `self.B`.

        F : np.array(dim_x, dim_x), or None
            Optional state transition matrix; a value of None
            will cause the filter to use `self.F`.

        Q : np.array(dim_x, dim_x), scalar, or None
            Optional process noise matrix; a value of None will cause the
            filter to use `self.Q`.
        """

        if B is None:
            B = self.B
        if F is None:
            F = self.F
        if Q is None:
            Q = self.Q
        elif isscalar(Q):
            Q = eye(self.dim_x) * Q


        # x = Fx + Bu
        if blindSpot == None:
            if B is not None and u is not None:
                self.x = dot(F, self.x) + dot(B, u)
            else:
                self.x = dot(F, self.x)

            # P = FPF' + Q
            self.P = self._alpha_sq * dot(dot(F, self.P), F.T) + Q

            # save prior
            self.x_prior = self.x.copy()
            self.P_prior = self.P.copy()

        elif blindSpot == True:

            self.x = dot(self.F_blindspot, self.x)

            # P = FPF' + Q
            self.P = self._alpha_sq * dot(dot(self.F_blindspot, self.P), self.F_blindspot.T) + Q

            # save prior
            self.x_prior = self.x.copy()
            self.P_prior = self.P.copy()

        if self.x[3] >= np.pi: self.x[3] -= np.pi * 2
        if self.x[3] < -np.pi: self.x[3] += np.pi * 2

    def update(self, z, R=None, H=None):

        """
        bs = 1 
        Zs: measurement                                                              (bs, no)
        Xs: estimate                                                                 (bs, ns)
        Ps: uncertainty covariance                                                   (bs, ns * ns)
        self.x :                                                                     (ns, 1)
        self.P :                                                                     (ns, ns)
        """


        bs = 1
        ns = 10
        no = 7

        z = z.astype(self.dt)
        self.x = self.x.astype(self.dt)
        self.P = self.P.astype(self.dt)

        Zs = z.reshape(bs, no)
        Xs = self.x.reshape(bs, ns)
        Ps = self.P.reshape(bs, ns * ns)

        kalman_update_batch_gpu(Zs, Xs, Ps, bs, ns, no)

        self.x = Xs.reshape(ns, 1)
        self.P = Ps.reshape(ns, ns)

    








        
