

import numpy as np
from scipy.optimize import linear_sum_assignment
from AB3DMOT_libs.bbox_utils import convert_3dbox_to_8corner, iou3d
from AB3DMOT_libs.kalman_filter import KalmanBoxTracker
import time
from gpu_helpers import iou2d_gpu, iou3d_gpu, kalman_update_batch_gpu
from Queue import deque


CLASS_CYCLIST = 5


def distance(xy1, xy2):
    return ((xy1[0] - xy2[0]) ** 2 + (xy1[1] - xy2[1]) ** 2) ** 0.5 

def in_blind(xy, area):

    x_min, x_max, y_min, y_max = area
    if (x_min < xy[0] < x_max) & (y_min < xy[1] < y_max):
        return True
    else:
        return False

def blind_spot_in(unmatched_trks, trackers, area_in, blind_trks, maxbs):
    blind_ids = []
    todel = []
    if len(blind_trks) != 0:
        blind_ids = [trk.id for trk in blind_trks]
    for t, trk in enumerate(trackers):
        if t in unmatched_trks:
            xypos = trk.kf.x[:2]
            if in_blind(xypos, area_in): 
                trk.time_since_update = 0 # simulate an update, otherwise this trk will be deleted at the end
                trk.hits += 1             # simulate an update
                trk.blind_stay += 1

                # one more small predication for blind trks, accelerate the motion
                trk.kf.predict(blindSpot=True)           

                # avoid repeating adding the same trk and reject too-long blind trk
                if trk.blind_stay < maxbs and trk.id not in blind_ids:  
                    blind_trks.append(trk)
                
                # delete elements that have stayed 20 updates in blind area
                if trk.blind_stay >= maxbs and trk.id in blind_ids: 
                    todel.append(t)

    trackers = [trackers[i] for i in range(len(trackers)) if i not in todel]
    return trackers

def blind_spot_out(unmatched_dets, dets, blind_trks, area_out, info, area_in):
  
    # dets:[x,y,z,theta,l,w,h]
    if len(blind_trks) == 0:
        return [], []

    assigned = []
    matched_dets_spot = []

    for i in unmatched_dets:
        xypos = dets[i, :2]
        if in_blind(xypos, area_in) and int(info[i, 1]) == CLASS_CYCLIST:
            matched_dets_spot.append(i)   # false alarm for cyclist, ignore
            continue

        if in_blind(xypos, area_out):
            min_dist = np.inf
            idx = None       # matched trk idx for this det

            for t, trk in enumerate(blind_trks):
                if t in assigned:
                    continue
                dist = distance(trk.kf.x[:2], xypos)        
                if dist < min_dist:
                    min_dist = dist
                    idx = t

            if idx == None:
                continue

            blind_trks[idx].update_single(dets[i], trk.info)
            blind_trks[idx].kf.predict()
            assigned.append(idx)
            matched_dets_spot.append(i)

    blind_trks = deque([trk for t, trk in enumerate(blind_trks) if t not in assigned]) # remove assigned blind-trks, keep it in self.trackers!

    return matched_dets_spot, blind_trks
    
def batchwise_kalman_update(dets, trackers, unmatched_trks, matched, info):
    
    ttm = time.time()
    bs = 0
    ts, ds, nts = [], [], []
    Xs, Ps, Zs = [], [], []

    dt = np.float32
    ns = 10
    no = 7
    for t, trk in enumerate(trackers):
        if t not in unmatched_trks:
            d = matched[np.where(matched[:, 1] == t)[0], 0]  # a list of index
            bbox, _,  _ = trk.correct_angle(dets[d, :][0])
            dets[d, :] = bbox

            Zs.append(bbox.reshape(1, no))
            Xs.append(trk.kf.x.reshape(1, ns))
            Ps.append(trk.kf.P.reshape(1, ns, ns))

            ts.append(t)  
            ds.append(d)
            bs += 1
        else:
            nts.append(t)

    assert len(ds) == len(ts) == bs == len(trackers) - len(nts)
    assert len(Zs) == len(Xs) == len(Ps) == bs

    if bs != 0:
        Zs = np.concatenate(Zs).astype(dt)
        Xs = np.concatenate(Xs).astype(dt)
        Ps = np.concatenate(Ps).astype(dt)
        Ps = Ps.reshape(bs, ns * ns)
    else:
        Xs = np.zeros((bs, ns)).astype(dt)
        Ps = np.zeros((bs, ns * ns)).astype(dt)
        Zs = np.zeros((bs, no)).astype(dt)


    t3_bch_s1 = time.time() - ttm

    ttm = time.time()
    kalman_update_batch_gpu(Zs, Xs, Ps, bs, ns, no)
    t3_bch_s2 = time.time() - ttm

    ttm = time.time()
    for i in range(bs):
        t = ts[i]
        d = ds[i]
        trk = trackers[t]
        trk.kf.x = Xs[i].reshape(ns, 1)
        trk.kf.P = Ps[i].reshape(ns, ns)

        ori_info = trk.info
        cur_info = info[d, :][0]
        if cur_info[1] == ori_info[1]:
            if ori_info[-1] !='None':
                cur_info[-1] = ori_info[-1]
            if ori_info[-2] !='None':
                cur_info[-2] = ori_info[-2]
            if ori_info[-3] !='None':
                cur_info[-3] = ori_info[-3]
    
        trk.assign_info(cur_info)

    for nt in nts:
        trackers[nt].info[0] = info[0][0]

    t3_bch_s3 = time.time() - ttm
    return t3_bch_s1, t3_bch_s2, t3_bch_s3

def associate_detections_to_trackers(detections, trackers, iou_threshold=0.01, timing_log_path=None):

    if timing_log_path is not None:
        log_file = open(timing_log_path, 'a')

    # TIMING PART 1
    s = time.time()

    if (len(trackers) == 0):
        return np.empty((0, 2), dtype=int), np.arange(len(detections)), np.empty((0, 8, 3), dtype=int)

    assert detections.shape[1] == trackers.shape[1] == 7


    iou_res = np.zeros((len(detections), len(trackers)), dtype=np.float32)

    t2_0 = time.time() - s

    # TIMING PART 2
    s = time.time()
    
    
    # here det and trk are N * 7, not N * 8 * 3, [x,y,z,l,w,h,thetha]
    iou_res = iou2d_gpu(detections, trackers)   # m * n
    # order conversion:
    gpu_reorder_b = [0, 1, 2, 6, 3, 4, 5]
    detections = detections[:, gpu_reorder_b]
    trackers = trackers[:, gpu_reorder_b]

    t2_1 = time.time() - s
    # TIMING PART 3
    s = time.time()

    row_ind, col_ind = linear_sum_assignment(-iou_res)  # hougarian algorithm
    matched_indices = np.stack((row_ind, col_ind), axis=1)

    t2_2 = time.time() - s
    # TIMING PART 4
    s = time.time()

    unmatched_detections = []
    for d, det in enumerate(detections):
        if (d not in matched_indices[:, 0]): unmatched_detections.append(d)
    unmatched_trackers = []
    for t, trk in enumerate(trackers):
        if (t not in matched_indices[:, 1]): unmatched_trackers.append(t)

    # filter out matched with low IOU
    matches = []
    for m in matched_indices:
        if (iou_res[m[0], m[1]] < iou_threshold):
            unmatched_detections.append(m[0])
            unmatched_trackers.append(m[1])
        else:
            matches.append(m.reshape(1, 2))
    if (len(matches) == 0):
        matches = np.empty((0, 2), dtype=int)
    else:
        matches = np.concatenate(matches, axis=0)

    t2_3 = time.time() - s

    '''
    if timing_log_path is not None:
        log_file.write("associate_detections_to_trackers: t2_0: %.3f ms, t2_1 IOU : %.3f ms, t2_2 Houngarian: %.3f ms, t2_3 Match: %.3f ms\n" % (t2_0*1000, t2_1*1000, t2_2*1000, t2_3*1000))
    '''

    return matches, np.array(unmatched_detections), np.array(unmatched_trackers)

class AB3DMOT(object):  # A baseline of 3D multi-object tracking
    # max age will preserve the bbox does not appear no more than 2 frames, interpolate the detection
    def __init__(self, max_age=2,
                 min_hits=3, blind_in=None, blind_out=None):
        """
        Sets key parameters for SORT
        """
        self.max_age = max_age
        self.min_hits = min_hits
        self.trackers = []
        self.frame_count = 0
        self.reorder = [3, 4, 5, 6, 2, 1, 0]
        self.reorder_back = [6, 5, 4, 0, 1, 2, 3]

        #area: [x_min, x_max, y_min, y_max]
        # N1_blind_in = [0, 8, -8, 9]
        # N1_blind_out = [0, 8, 6, 15]
        # N7_blind_in = [0, 8, -8, 9]
        # N7_blind_out = [0, 8, -15, -7]
        assert blind_in is not None
        assert blind_out is not None
        
        self.blind_in = blind_in
        self.blind_out = blind_out

        self.blind_trks = deque([])
        self.max_blind_trks = 4
        self.max_blind_stay = 30


    def blind_spot(self, unmatched_trks, unmatched_dets, dets, info):

        if len(self.blind_in) == len(self.blind_out) == 0:
            return [], self.blind_trks
            
        while len(self.blind_trks) > self.max_blind_trks:
            self.blind_trks.popleft()
        
        ls = []
        all_ids = [trk.id for trk in self.trackers]
        for tb, trkb in enumerate(self.blind_trks):         # make sure blind_trks only contain ongoing trks
            if trkb.id in all_ids:
                ls.append(tb)
        self.blind_trks = deque([self.blind_trks[tb] for tb in ls])

        self.trackers = blind_spot_in(unmatched_trks, self.trackers, self.blind_in, self.blind_trks, self.max_blind_stay)
        assigned_det_blind, self.blind_trks = blind_spot_out(unmatched_dets, dets, self.blind_trks, self.blind_out, info, self.blind_in)
        return assigned_det_blind, self.blind_trks

    def update(self, dets_all, timing_log_path=None):
        """
        Params:
          dets_all: dict
            dets - a numpy array of detections in the format [[h,w,l,x,y,z,theta],...]
            info: a array of other info for each det
        Requires: this method must be called once for each frame even with empty detections.
        Returns the a similar array, where the last column is the object ID.

        NOTE: The number of objects returned may differ from the number of detections provided.
        """

        gpu_iou = True
        
        if timing_log_path is not None:
            log_file = open(timing_log_path, 'a')


        # TIMING PART 1
        s = time.time()
        
        dets, info = dets_all['dets'], dets_all['info']  # dets: N x 7, float numpy array

        # reorder the data to put x,y,z in front to be compatible with the state transition matrix
        # where the constant velocity model is defined in the first three rows of the matrix
        dets = dets[:, self.reorder]                                                    # reorder the data to [[x,y,z,theta,l,w,h], ...]
        self.frame_count += 1

        trks = np.zeros((len(self.trackers), 7))  # N x 7 ,                             # get predicted locations from existing trackers.
        to_del = []                                                                     # the tracking to be deleted
        ret = []                                                                        # results
        for t, trk in enumerate(trks):                                                  # 
            pos = self.trackers[t].predict().reshape((-1, 1))
            trk[:] = [pos[0], pos[1], pos[2], pos[3], pos[4], pos[5], pos[6]]
            if (np.any(np.isnan(pos))):
                to_del.append(t)
        trks = np.ma.compress_rows(np.ma.masked_invalid(trks))
        for t in reversed(to_del):
            self.trackers.pop(t)

        t0 = time.time() - s
        # TIMING PART 2
        s = time.time()   
        
        t1 = time.time() - s
        # TIMING PART 3
        s = time.time()
        
        # dets: [x,y,z,theta,l,w,h] 
        gpu_reorder = [0, 1, 2, 4, 5, 6, 3] 
        dets_n7 = np.zeros_like(dets)
        trks_n7 = np.zeros_like(trks)
        dets_n7 = dets[:, gpu_reorder]
        trks_n7 = trks[:, gpu_reorder]

        # inputs are N * 7 [x,y,z,l,w,h,thetha]
        matched, unmatched_dets, unmatched_trks = associate_detections_to_trackers(dets_n7, trks_n7, timing_log_path=timing_log_path)  # Hungarian algorithm for matching
    

        t2 = time.time() - s
        # TIMING PART 4
        s = time.time()
        bboxnum = 0

        t3_bch_s1, t3_bch_s2, t3_bch_s3 = batchwise_kalman_update(dets, self.trackers, unmatched_trks, matched, info)
    
        num_t = len(self.trackers)

        t3 = time.time() - s
        # TIMING PART 5
        s = time.time()

        assigned_det_blind, self.blind_trks = self.blind_spot(unmatched_trks, unmatched_dets, dets, info)

        t_blind = time.time() - s

        s = time.time()
        # create and initialise new trackers for unmatched detections
        for i in unmatched_dets:  # a scalar of index
            if i in assigned_det_blind:  
                continue
            trk = KalmanBoxTracker(dets[i, :], info[i, :])
            self.trackers.append(trk)

        t4 = time.time() - s
        # TIMING PART 6
        s = time.time()
    
        """ 
        if timing_log_path is not None:
            log_file.write("trackers id: \n")
            for i, t in enumerate(self.trackers):
                d = t.get_state()
                xy = d[:2]
                it = in_blind(xy, self.blind_in)
                blindstay = t.blind_stay

                log_file.write("id: %d, typ: %d, blind: %s blindstay: %d- \n" % (t.id, int(t.info[1]), it, blindstay))

            log_file.write("\nblind trackers id: \n")

            for k in self.blind_trks:
                i = k[0]
                log_file.write("len(self.trackers): %d, i: %d, id:%d- \n" % (len(self.trackers),i, k[2]))
            log_file.write('unmatched dets: \n')
            for i in unmatched_dets: 
                log_file.write("type: %d\n" % int(info[i, 1]))
            log_file.write("*"*10)

        """

        i = len(self.trackers)
        for trk in reversed(self.trackers):
            d = trk.get_state()  # bbox location
            d = d[self.reorder_back]  # change format from [x,y,z,theta,l,w,h] to [h,w,l,x,y,z,theta]

            orientation = trk.get_orientation()

            if ((trk.time_since_update < self.max_age) and (
                    trk.hits >= self.min_hits or self.frame_count <= self.min_hits)):

                ret.append(np.concatenate((d, [trk.id + 1], orientation, trk.info, [trk.create_stamp, len(trk.last_x_y_list[:])])).reshape(1, -1))  # +1 as MOT benchmark requires positive
            i -= 1

            # remove dead tracklet,
            # means this trk has predicted twice after one update  ===> dead
            if (trk.time_since_update >= self.max_age):
                self.trackers.pop(i)
    
        t5 = time.time() - s

        '''
        if timing_log_path is not None:
            log_file.write("AB3DMOT.update(): t0 predict:%.3f ms, t1: %.3f ms, t2 iou and hougarian: %.3f ms, t3 kalman: %.3f ms, \
                        t3_bch_s1 angle,dataPrepare: %.3f ms, t3_bch_s2 kalmanOP: %.3f ms, t3_bch_s3 scatter&post: %.3f ms, \
                        t_blind: %.3f ms,  t4 new tracker: %.3f ms, t5 tracker addremove: %.3f ms, bbox num: %d, tracker num: %d\
                         len(blind_tracks) %d \n" % (t0*1000, t1*1000, t2*1000, t3*1000, t3_bch_s1*1000, t3_bch_s2*1000,t3_bch_s3*1000,
                         t_blind * 1000, t4*1000, t5*1000, bboxnum, num_t, len(self.blind_trks)))
        '''

        if (len(ret) > 0): 
            return np.concatenate(ret)  # h,w,l,x,y,z,theta, ID, other info, confidence
        return np.empty((0, 19))



    def predict(self, timestamp):

        rets = []

        if len(self.trackers) == 0:
            return rets

        pre_timestamp = int(self.trackers[0].info[0])
        if (timestamp - pre_timestamp) > 450:
            return rets

        count = int(round((timestamp - pre_timestamp) / 100.0))
        index = 1
        while index < count:
            self.frame_count += 1

            trks = np.zeros((len(self.trackers), 7))
            to_del = []
            ret = []
            for t, trk in enumerate(trks):
                pos = self.trackers[t].predict().reshape((-1, 1))
                if (np.any(np.isnan(pos))):
                    to_del.append(t)

            for t in reversed(to_del):
                self.trackers.pop(t)

            for trk in reversed(self.trackers):
                d = trk.get_state()  # bbox location
                d = d[self.reorder_back]  # change format from [x,y,z,theta,l,w,h] to [h,w,l,x,y,z,theta]

                orientation = trk.get_orientation()

                trk.info[0] = pre_timestamp + 100*index
                ret.append(np.concatenate((d, [trk.id + 1], orientation, trk.info, [trk.create_stamp, len(trk.last_x_y_list[:])])).reshape(1, -1))  # +1 as MOT benchmark requires positive

            rets.append(np.concatenate(ret))
            index += 1
        return rets

