
#include "common.hpp"

#define CHECK(status)                                          \
    do {                                                       \
        auto ret = (status);                                   \
        if (ret != 0) {                                        \
            std::cerr << "Cuda failure: " << ret << std::endl; \
            abort();                                           \
        }                                                      \
    } while (0)

using namespace nvinfer1;

cv::Mat preprocess_img(cv::Mat& img) {
    int w, h, x, y;
    float r_w = Yolo::INPUT_W / (img.cols * 1.0);
    float r_h = Yolo::INPUT_H / (img.rows * 1.0);
    if (r_h > r_w) {
        w = Yolo::INPUT_W;
        h = r_w * img.rows;
        x = 0;
        y = (Yolo::INPUT_H - h) / 2;
    } else {
        w = r_h * img.cols;
        h = Yolo::INPUT_H;
        x = (Yolo::INPUT_W - w) / 2;
        y = 0;
    }
    cv::Mat re(h, w, CV_8UC3);
    cv::resize(img, re, re.size(), 0, 0, cv::INTER_LINEAR);
    cv::Mat out(Yolo::INPUT_H, Yolo::INPUT_W, CV_8UC3, cv::Scalar(128, 128, 128));
    re.copyTo(out(cv::Rect(x, y, re.cols, re.rows)));

    // cv::Mat  out;
    // cv::resize(img,out,cv::Size(Yolo::INPUT_H,Yolo::INPUT_W));
    return out;
}

cv::Rect get_rect(cv::Mat& img, float bbox[4]) {
    int l, r, t, b;
    float r_w = Yolo::INPUT_W / (img.cols * 1.0);
    float r_h = Yolo::INPUT_H / (img.rows * 1.0);
    if (r_h > r_w) {
        l = bbox[0] - bbox[2] / 2.f;
        r = bbox[0] + bbox[2] / 2.f;
        t = bbox[1] - bbox[3] / 2.f - (Yolo::INPUT_H - r_w * img.rows) / 2;
        b = bbox[1] + bbox[3] / 2.f - (Yolo::INPUT_H - r_w * img.rows) / 2;
        l = l / r_w;
        r = r / r_w;
        t = t / r_w;
        b = b / r_w;
    } else {
        l = bbox[0] - bbox[2] / 2.f - (Yolo::INPUT_W - r_h * img.cols) / 2;
        r = bbox[0] + bbox[2] / 2.f - (Yolo::INPUT_W - r_h * img.cols) / 2;
        t = bbox[1] - bbox[3] / 2.f;
        b = bbox[1] + bbox[3] / 2.f;
        l = l / r_h;
        r = r / r_h;
        t = t / r_h;
        b = b / r_h;
    }
    // l = bbox[0] - bbox[2] / 2.f;
    // r = bbox[0] + bbox[2] / 2.f;
    // t = bbox[1] - bbox[3] / 2.f;
    // b = bbox[1] + bbox[3] / 2.f;
    // l = l / r_w;
    // r = r / r_w;
    // t = t / r_h;
    // b = b / r_h;
    if(l < 0){
        l = 0;
    }
    if(t < 0){
        t = 0;
    }
    if(r > img.cols){
        r = img.cols;
    }
    if(b > img.rows){
        b = img.rows;
    }
    return cv::Rect(l, t, r - l, b - t);
}

float iou(float lbox[4], float rbox[4]) {
    float interBox[] = {
        (std::max)(lbox[0] - lbox[2] / 2.f, rbox[0] - rbox[2] / 2.f),  // left
        (std::min)(lbox[0] + lbox[2] / 2.f, rbox[0] + rbox[2] / 2.f),  // right
        (std::max)(lbox[1] - lbox[3] / 2.f, rbox[1] - rbox[3] / 2.f),  // top
        (std::min)(lbox[1] + lbox[3] / 2.f, rbox[1] + rbox[3] / 2.f),  // bottom
    };

    if (interBox[2] > interBox[3] || interBox[0] > interBox[1]) return 0.0f;

    float interBoxS = (interBox[1] - interBox[0]) * (interBox[3] - interBox[2]);
    return interBoxS / (lbox[2] * lbox[3] + rbox[2] * rbox[3] - interBoxS);
}

bool cmp(const Yolo::Detection& a, const Yolo::Detection& b) { return a.conf > b.conf; }

void nms(std::vector<Yolo::Detection>& res, float* output, float conf_thresh, float nms_thresh /* = 0.5*/) {
    int det_size = sizeof(Yolo::Detection) / sizeof(float);
    std::map<float, std::vector<Yolo::Detection>> m;
    for (int i = 0; i < output[0] && i < Yolo::MAX_OUTPUT_BBOX_COUNT; i++) {
        if (output[1 + det_size * i + 4] <= conf_thresh) continue;
        Yolo::Detection det;
        memcpy(&det, &output[1 + det_size * i], det_size * sizeof(float));
        if (m.count(det.class_id) == 0) m.emplace(det.class_id, std::vector<Yolo::Detection>());
        m[det.class_id].push_back(det);
    }
    for (auto it = m.begin(); it != m.end(); it++) {
        // std::cout << it->second[0].class_id << " --- " << std::endl;
        auto& dets = it->second;
        std::sort(dets.begin(), dets.end(), cmp);
        for (size_t m = 0; m < dets.size(); ++m) {
            auto& item = dets[m];
            res.push_back(item);
            for (size_t n = m + 1; n < dets.size(); ++n) {
                if (iou(item.bbox, dets[n].bbox) > nms_thresh) {
                    dets.erase(dets.begin() + n);
                    --n;
                }
            }
        }
    }
}

void reducerepeat(std::vector<Yolo::Detection>& res, float nms_thresh){  //去掉重复框
    for(int i = 0; i < res.size(); i++){
        auto item = res[i];
        for(int j = i + 1; j < res.size(); j++){
            auto nextitem = res[j];
            float ioudata = iou(item.bbox,nextitem.bbox);
            if(ioudata > 0.8){
                if(item.conf > nextitem.conf){
                    res.erase(res.begin() + j);
                    j--;
                }
                else{
                    res.erase(res.begin() + i);
                    i--;
                    break;
                }
            }
        }
    }
}

void checkpeoplelocation(std::vector<Yolo::Detection>& res){  //如果人在车内,则删除人
    for(int i = 0;i < res.size(); i++){
        auto item = res[i];
        if(item.class_id != 0){
            continue;
        }
        else{
            for (int j = 0; j < res.size(); j++){
                if(j == i){
                    continue;
                }
                else{
                    auto &item2 = res[j];
                    //判断人和另一个物体有没有交集
                    float ioudata = iou(item.bbox,item2.bbox);
                    if(ioudata <= 0.0){
                        continue;
                    }


                    //item2坐标转换xywh->xyxy
                    int l = item2.bbox[0] - item2.bbox[2] / 2.0;
                    int t = item2.bbox[1] - item2.bbox[3] / 2.0;
                    int r = item2.bbox[0] + item2.bbox[2] / 2.0;
                    int b = item2.bbox[1] + item2.bbox[3] / 2.0;
                    if(l < 0){
                        l = 0;
                    }
                    if(t < 0){
                        t = 0;
                    }
                    if(r > Yolo::INPUT_W){
                        r = Yolo::INPUT_W;
                    }
                    if(b > Yolo::INPUT_H){
                        b = Yolo::INPUT_H;
                    }

                    //item坐标转换xywh->xyxy
                    int l1 = item.bbox[0] - item.bbox[2] / 2.0;
                    int t1 = item.bbox[1] - item.bbox[3] / 2.0;
                    int r1 = item.bbox[0] + item.bbox[2] / 2.0;
                    int b1 = item.bbox[1] + item.bbox[3] / 2.0;
                    if(l1 < 0){
                        l1 = 0;
                    }
                    if(t1 < 0){
                        t1 = 0;
                    }
                    if(r1 > Yolo::INPUT_W){
                        r1 = Yolo::INPUT_W;
                    }
                    if(b1 > Yolo::INPUT_H){
                        b1 = Yolo::INPUT_H;
                    }

                    if(l <= l1 && t <= t1 && r >= r1 && b >= b1){
                        res.erase(res.begin() + i);
                        i--;
                        break;
                    }
               }
            }
        }
    }
}

void mergeperson(std::vector<Yolo::Detection>& res){ //对于人和自行车,人和摩托车,做框的合并
    for(int i = 0; i < res.size(); i++){
        auto item = res[i];
        if(item.class_id != 0){
            continue;
        }
        else{
            for(int j = 0; j < res.size();j++){
                if(j == i){
                    continue;
                }
                else{
                    auto &item2 = res[j];
                    if(item2.class_id != 2 && item2.class_id != 3){
                        continue;
                    }
                    //判断人和另一个物体有没有交集
                    float ioudata = iou(item.bbox,item2.bbox);
                    if(ioudata >= 0.2){
                        int minx = std::min(item.bbox[0] - item.bbox[2] / 2.0,item2.bbox[0] - item2.bbox[2] / 2.0);
                        int miny = std::min(item.bbox[1] - item.bbox[3] / 2.0,item2.bbox[1] - item2.bbox[3] / 2.0);
                        int maxx = std::max(item.bbox[0] + item.bbox[2] / 2.0,item2.bbox[0] + item2.bbox[2] / 2.0);
                        int maxy = std::max(item.bbox[1] + item.bbox[3] / 2.0,item2.bbox[1] + item2.bbox[3] / 2.0);
                        res[j].bbox[0] = (minx + maxx) / 2.0;
                        res[j].bbox[1] = (miny + maxy) / 2.0;
                        res[j].bbox[2] = maxx - minx;
                        res[j].bbox[3] = maxy - miny;

                        res.erase(res.begin() + i);
                        i--;
                        break;
                    }
                }
            }                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          
        }
    }
}


std::string recognizeVehicleColor1(const cv::Mat &img)
{
    cv::Size imgSize = img.size();
    cv::Mat preImg = img(cv::Rect(imgSize.width * 0.15, imgSize.height * 0.15, imgSize.width * (1 - 0.3), imgSize.height * (1 - 0.3)));
    // HSV 三通道操作
    // imshow("preImg", preImg);
    //cv::waitKey();
    cv::Mat hsvImg;
    cv::cvtColor(preImg, hsvImg, CV_BGR2HSV);
    std::vector<cv::Mat> hsv_planes;
    cv::split(hsvImg, hsv_planes); // 分离HSV通道
    cv::Mat hMask, sMask, vMask;

    cv::threshold(hsv_planes[1], sMask, 43, 255, cv::THRESH_BINARY); // 以43为阈值将S通道进行二值化  43来自 HSV基本颜色分量范围
    //cv::multiply(sMask, bgrMask / 255, hMask);						// 与背景 Mask 合成为一个 Mask

    cv::threshold(hsv_planes[2], vMask, 46, 255, cv::THRESH_BINARY); // 以46为阈值将V通道进行二值化  46来自 HSV基本颜色分量范围
    cv::multiply(vMask, sMask / 255, hMask);                         // 与背景 Mask 合成为一个 Mask 用于H通道

    cv::threshold(hsv_planes[1], vMask, 43, 255, cv::THRESH_BINARY_INV); // 以43为阈值将S通道进行二值化 用于V通道

    // 利用二值图 Mask 将 HSV 通道中的背景部分设为 0.
    hsv_planes[1] = hsv_planes[2].clone();
    cv::multiply(hsv_planes[0], hMask / 255, hsv_planes[0]);
    cv::multiply(hsv_planes[2], vMask / 255, hsv_planes[2]);

    // 分别计算三个通道的颜色直方图
    int histSize = 180; // 将色调量化到180个级别 对应 H 取值范围 [0:180]
    float ranges[] = {0, 180};
    const float *hRanges = {ranges};
    cv::Mat hsv_hist, sv_hist, v_hist;
    //计算直方图
    calcHist(&hsv_planes[0], 1, 0, cv::Mat(), hsv_hist, 1, &histSize, &hRanges, true, false);
    histSize = 256;
    ranges[1] = 256;
    const float *Ranges = {ranges};
    calcHist(&hsv_planes[2], 1, 0, cv::Mat(), sv_hist, 1, &histSize, &Ranges, true, false);
    calcHist(&hsv_planes[1], 1, 0, cv::Mat(), v_hist, 1, &histSize, &Ranges, true, false);
    std::map<std::string, float> color; // 定义颜色映射， 对每个颜色对应的HSV区间内进行统计
    color["red"] = 0;
    color["orange"] = 0;
    color["yellow"] = 0;
    color["green"] = 0;
    color["cyan"] = 0;
    color["blue"] = 0;
    color["violet"] = 0;
    color["black"] = 0;
    color["grey"] = 0;
    color["white"] = 0;
    for (int r = 0; r < hsv_hist.rows; r++)
    {
        float binVal = hsv_hist.at<float>(r);
        if ((r <= 10 && r > 0) || (r >= 160)) // 红色区间 H [0,10]
        {
            color["red"] += binVal;
        }
        else if (r >= 11 && r <= 25) // 橙色区间 H [11,25]
        {
            color["orange"] += binVal;
        }
        else if (r >= 26 && r <= 34) // 黄色区间 H [26,34]
        {
            color["yellow"] += binVal;
        }
        else if (r >= 35 && r <= 77) // 绿色区间 H [35,77]
        {
            color["green"] += binVal;
        }
        else if (r >= 78 && r <= 99) // 青色区间 H [78,99]
        {
            color["cyan"] += binVal;
        }
        else if (r >= 100 && r <= 124) // 蓝色区间 H [100,124]
        {
            color["blue"] += binVal;
        }
        else if (r >= 125 && r <= 159) // 紫色区间 H [125,155]
        {
            color["violet"] += binVal;
        }
    }
    for (int r = 1; r < sv_hist.rows; r++)
    {
        float sv_binVal = sv_hist.at<float>(r);
        if (r > 200)
        {
            color["white"] += sv_binVal; // 白色区间 H [200,255]
        }
    }
    for (int r = 1; r < v_hist.rows; r++)
    {
        float v_binVal = v_hist.at<float>(r);
        if (r > 0 && r <= 46)
        {
            color["black"] += v_binVal; // 黑色区间 H [0,40]
        }
    }
    float maxVal = 0;
    std::string result = "";
    std::map<std::string, float>::const_iterator it = color.begin(); //找出对应颜色区间统计数最多的颜色
    for (it; it != color.end(); it++)
    {
        if (it->second > maxVal)
        {
            maxVal = it->second;
            result = it->first;
        }
    }
    // 其他颜色需要大于白色的2倍
    if (((color["white"] / color[result]) > 0.5) &&  result != "white")
    {
        result = "white";
    }
    return result;
}




// TensorRT weight files have a simple space delimited format:
// [type] [size] <data x size in hex>
std::map<std::string, Weights> loadWeights(const std::string file) {
    std::cout << "Loading weights: " << file << std::endl;
    std::map<std::string, Weights> weightMap;

    // Open weights file
    std::ifstream input(file);
    assert(input.is_open() && "Unable to load weight file. please check if the .wts file path is right!!!!!!");

    // Read number of weight blobs
    int32_t count;
    input >> count;
    assert(count > 0 && "Invalid weight map file.");

    while (count--) {
        Weights wt{DataType::kFLOAT, nullptr, 0};
        uint32_t size;

        // Read name and type of blob
        std::string name;
        input >> name >> std::dec >> size;
        wt.type = DataType::kFLOAT;

        // Load blob
        uint32_t* val = reinterpret_cast<uint32_t*>(malloc(sizeof(val) * size));
        for (uint32_t x = 0, y = size; x < y; ++x) {
            input >> std::hex >> val[x];
        }
        wt.values = val;

        wt.count = size;
        weightMap[name] = wt;
    }

    return weightMap;
}

IScaleLayer* addBatchNorm2d(INetworkDefinition* network, std::map<std::string, Weights>& weightMap, ITensor& input, std::string lname, float eps) {
    float* gamma = (float*)weightMap[lname + ".weight"].values;
    float* beta = (float*)weightMap[lname + ".bias"].values;
    float* mean = (float*)weightMap[lname + ".running_mean"].values;
    float* var = (float*)weightMap[lname + ".running_var"].values;
    int len = weightMap[lname + ".running_var"].count;

    float* scval = reinterpret_cast<float*>(malloc(sizeof(float) * len));
    for (int i = 0; i < len; i++) {
        scval[i] = gamma[i] / sqrt(var[i] + eps);
    }
    Weights scale{DataType::kFLOAT, scval, len};

    float* shval = reinterpret_cast<float*>(malloc(sizeof(float) * len));
    for (int i = 0; i < len; i++) {
        shval[i] = beta[i] - mean[i] * gamma[i] / sqrt(var[i] + eps);
    }
    Weights shift{DataType::kFLOAT, shval, len};

    float* pval = reinterpret_cast<float*>(malloc(sizeof(float) * len));
    for (int i = 0; i < len; i++) {
        pval[i] = 1.0;
    }
    Weights power{DataType::kFLOAT, pval, len};

    weightMap[lname + ".scale"] = scale;
    weightMap[lname + ".shift"] = shift;
    weightMap[lname + ".power"] = power;
    IScaleLayer* scale_1 = network->addScale(input, ScaleMode::kCHANNEL, shift, scale, power);
    assert(scale_1);
    return scale_1;
}

ILayer* convBlock(INetworkDefinition* network, std::map<std::string, Weights>& weightMap, ITensor& input, int outch, int ksize, int s, int g, std::string lname) {
    Weights emptywts{DataType::kFLOAT, nullptr, 0};
    int p = ksize / 2;
    IConvolutionLayer* conv1 = network->addConvolutionNd(input, outch, DimsHW{ksize, ksize}, weightMap[lname + ".conv.weight"], emptywts);
    assert(conv1);
    conv1->setStrideNd(DimsHW{s, s});
    conv1->setPaddingNd(DimsHW{p, p});
    conv1->setNbGroups(g);
    IScaleLayer* bn1 = addBatchNorm2d(network, weightMap, *conv1->getOutput(0), lname + ".bn", 1e-3);

    // hard_swish = x * hard_sigmoid
    auto hsig = network->addActivation(*bn1->getOutput(0), ActivationType::kHARD_SIGMOID);
    assert(hsig);
    hsig->setAlpha(1.0 / 6.0);
    hsig->setBeta(0.5);
    auto ew = network->addElementWise(*bn1->getOutput(0), *hsig->getOutput(0), ElementWiseOperation::kPROD);
    assert(ew);
    return ew;
}

ILayer* focus(INetworkDefinition* network, std::map<std::string, Weights>& weightMap, ITensor& input, int inch, int outch, int ksize, std::string lname) {
    ISliceLayer* s1 = network->addSlice(input, Dims3{0, 0, 0}, Dims3{inch, Yolo::INPUT_H / 2, Yolo::INPUT_W / 2}, Dims3{1, 2, 2});
    ISliceLayer* s2 = network->addSlice(input, Dims3{0, 1, 0}, Dims3{inch, Yolo::INPUT_H / 2, Yolo::INPUT_W / 2}, Dims3{1, 2, 2});
    ISliceLayer* s3 = network->addSlice(input, Dims3{0, 0, 1}, Dims3{inch, Yolo::INPUT_H / 2, Yolo::INPUT_W / 2}, Dims3{1, 2, 2});
    ISliceLayer* s4 = network->addSlice(input, Dims3{0, 1, 1}, Dims3{inch, Yolo::INPUT_H / 2, Yolo::INPUT_W / 2}, Dims3{1, 2, 2});
    ITensor* inputTensors[] = {s1->getOutput(0), s2->getOutput(0), s3->getOutput(0), s4->getOutput(0)};
    auto cat = network->addConcatenation(inputTensors, 4);
    auto conv = convBlock(network, weightMap, *cat->getOutput(0), outch, ksize, 1, 1, lname + ".conv");
    return conv;
}

ILayer* bottleneck(INetworkDefinition* network, std::map<std::string, Weights>& weightMap, ITensor& input, int c1, int c2, bool shortcut, int g, float e, std::string lname) {
    auto cv1 = convBlock(network, weightMap, input, (int)((float)c2 * e), 1, 1, 1, lname + ".cv1");
    auto cv2 = convBlock(network, weightMap, *cv1->getOutput(0), c2, 3, 1, g, lname + ".cv2");
    if (shortcut && c1 == c2) {
        auto ew = network->addElementWise(input, *cv2->getOutput(0), ElementWiseOperation::kSUM);
        return ew;
    }
    return cv2;
}

ILayer* bottleneckCSP(INetworkDefinition* network, std::map<std::string, Weights>& weightMap, ITensor& input, int c1, int c2, int n, bool shortcut, int g, float e, std::string lname) {
    Weights emptywts{DataType::kFLOAT, nullptr, 0};
    int c_ = (int)((float)c2 * e);
    auto cv1 = convBlock(network, weightMap, input, c_, 1, 1, 1, lname + ".cv1");
    auto cv2 = network->addConvolutionNd(input, c_, DimsHW{1, 1}, weightMap[lname + ".cv2.weight"], emptywts);
    ITensor* y1 = cv1->getOutput(0);
    for (int i = 0; i < n; i++) {
        auto b = bottleneck(network, weightMap, *y1, c_, c_, shortcut, g, 1.0, lname + ".m." + std::to_string(i));
        y1 = b->getOutput(0);
    }
    auto cv3 = network->addConvolutionNd(*y1, c_, DimsHW{1, 1}, weightMap[lname + ".cv3.weight"], emptywts);

    ITensor* inputTensors[] = {cv3->getOutput(0), cv2->getOutput(0)};
    auto cat = network->addConcatenation(inputTensors, 2);

    IScaleLayer* bn = addBatchNorm2d(network, weightMap, *cat->getOutput(0), lname + ".bn", 1e-4);
    auto lr = network->addActivation(*bn->getOutput(0), ActivationType::kLEAKY_RELU);
    lr->setAlpha(0.1);

    auto cv4 = convBlock(network, weightMap, *lr->getOutput(0), c2, 1, 1, 1, lname + ".cv4");
    return cv4;
}

ILayer* SPP(INetworkDefinition* network, std::map<std::string, Weights>& weightMap, ITensor& input, int c1, int c2, int k1, int k2, int k3, std::string lname) {
    int c_ = c1 / 2;
    auto cv1 = convBlock(network, weightMap, input, c_, 1, 1, 1, lname + ".cv1");

    auto pool1 = network->addPoolingNd(*cv1->getOutput(0), PoolingType::kMAX, DimsHW{k1, k1});
    pool1->setPaddingNd(DimsHW{k1 / 2, k1 / 2});
    pool1->setStrideNd(DimsHW{1, 1});
    auto pool2 = network->addPoolingNd(*cv1->getOutput(0), PoolingType::kMAX, DimsHW{k2, k2});
    pool2->setPaddingNd(DimsHW{k2 / 2, k2 / 2});
    pool2->setStrideNd(DimsHW{1, 1});
    auto pool3 = network->addPoolingNd(*cv1->getOutput(0), PoolingType::kMAX, DimsHW{k3, k3});
    pool3->setPaddingNd(DimsHW{k3 / 2, k3 / 2});
    pool3->setStrideNd(DimsHW{1, 1});

    ITensor* inputTensors[] = {cv1->getOutput(0), pool1->getOutput(0), pool2->getOutput(0), pool3->getOutput(0)};
    auto cat = network->addConcatenation(inputTensors, 4);

    auto cv2 = convBlock(network, weightMap, *cat->getOutput(0), c2, 1, 1, 1, lname + ".cv2");
    return cv2;
}

int read_files_in_dir(const char* p_dir_name, std::vector<std::string>& file_names) {
    DIR* p_dir = opendir(p_dir_name);
    if (p_dir == nullptr) {
        return -1;
    }

    struct dirent* p_file = nullptr;
    while ((p_file = readdir(p_dir)) != nullptr) {
        if (strcmp(p_file->d_name, ".") != 0 && strcmp(p_file->d_name, "..") != 0) {
            // std::string cur_file_name(p_dir_name);
            // cur_file_name += "/";
            // cur_file_name += p_file->d_name;
            std::string cur_file_name(p_file->d_name);
            file_names.push_back(cur_file_name);
        }
    }

    closedir(p_dir);
    return 0;
}

std::vector<float> getAnchors(std::map<std::string, Weights>& weightMap) {
    std::vector<float> anchors_yolo;
    Weights Yolo_Anchors = weightMap["model.24.anchor_grid"];
    assert(Yolo_Anchors.count == 18);
    int each_yololayer_anchorsnum = Yolo_Anchors.count / 3;
    const float* tempAnchors = (const float*)(Yolo_Anchors.values);
    for (int i = 0; i < Yolo_Anchors.count; i++) {
        if (i < each_yololayer_anchorsnum) {
            anchors_yolo.push_back(const_cast<float*>(tempAnchors)[i]);
        }
        if ((i >= each_yololayer_anchorsnum) && (i < (2 * each_yololayer_anchorsnum))) {
            anchors_yolo.push_back(const_cast<float*>(tempAnchors)[i]);
        }
        if (i >= (2 * each_yololayer_anchorsnum)) {
            anchors_yolo.push_back(const_cast<float*>(tempAnchors)[i]);
        }
    }
    return anchors_yolo;
}

IPluginV2Layer* addYoLoLayer(INetworkDefinition* network, std::map<std::string, Weights>& weightMap, IConvolutionLayer* det0, IConvolutionLayer* det1, IConvolutionLayer* det2) {
    auto creator = getPluginRegistry()->getPluginCreator("YoloLayer_TRT", "1");
    std::vector<float> anchors_yolo = getAnchors(weightMap);
    PluginField pluginMultidata[4];
    int NetData[4];
    NetData[0] = Yolo::CLASS_NUM;
    NetData[1] = Yolo::INPUT_W;
    NetData[2] = Yolo::INPUT_H;
    NetData[3] = Yolo::MAX_OUTPUT_BBOX_COUNT;
    pluginMultidata[0].data = NetData;
    pluginMultidata[0].length = 3;
    pluginMultidata[0].name = "netdata";
    pluginMultidata[0].type = PluginFieldType::kFLOAT32;
    int scale[3] = {8, 16, 32};
    int plugindata[3][8];
    std::string names[3];
    for (int k = 1; k < 4; k++) {
        plugindata[k - 1][0] = Yolo::INPUT_W / scale[k - 1];
        plugindata[k - 1][1] = Yolo::INPUT_H / scale[k - 1];
        for (int i = 2; i < 8; i++) {
            plugindata[k - 1][i] = int(anchors_yolo[(k - 1) * 6 + i - 2]);
        }
        pluginMultidata[k].data = plugindata[k - 1];
        pluginMultidata[k].length = 8;
        names[k - 1] = "yolodata" + std::to_string(k);
        pluginMultidata[k].name = names[k - 1].c_str();
        pluginMultidata[k].type = PluginFieldType::kFLOAT32;
    }
    PluginFieldCollection pluginData;
    pluginData.nbFields = 4;
    pluginData.fields = pluginMultidata;
    IPluginV2* pluginObj = creator->createPlugin("yololayer", &pluginData);
    ITensor* inputTensors_yolo[] = {det2->getOutput(0), det1->getOutput(0), det0->getOutput(0)};
    auto yolo = network->addPluginV2(inputTensors_yolo, 3, *pluginObj);
    return yolo;
}
