#include "yolov5.hpp"
#include "logging.h"
#include "yololayer.h"
#include "common.hpp"

YoloV5::YoloV5(){
    //参数配置
    input_w_ = Yolo::INPUT_W;;
    input_h_ = Yolo::INPUT_H;;
    input_size_ = 3 * input_h_ * input_w_;
    output_size_ = Yolo::MAX_OUTPUT_BBOX_COUNT * sizeof(Yolo::Detection) / sizeof(float) + 1;

}

void YoloV5::Init(std::string enginepath,int batchSize){
    batchSize_ = batchSize;
    enginepath_ = enginepath;
    //反序列化engine
    cudaSetDevice(0);
    char *trtModelStream =  nullptr;
    size_t size = 0 ;
    std::ifstream file(enginepath_, std::ios::binary);
    if (file.good()) {
        file.seekg(0, file.end);
        size = file.tellg();
        file.seekg(0, file.beg);
        trtModelStream = new char[size];
        assert(trtModelStream);
        file.read(trtModelStream, size);
        file.close();
    }

    static Logger gLogger;
    runtime_ = createInferRuntime(gLogger);
    assert(runtime_ != nullptr);
    engine_ = runtime_->deserializeCudaEngine(trtModelStream, size);
    assert(engine_ != nullptr);
    context_ = engine_->createExecutionContext();
    assert(context_ != nullptr);
   

    int inputIndex = engine_->getBindingIndex("data");
    int outputIndex = engine_->getBindingIndex("prob");
 

    CHECK(cudaMalloc(&buffers[inputIndex], batchSize_ * input_size_ * sizeof(float)));
    CHECK(cudaMalloc(&buffers[outputIndex], batchSize_ * output_size_ * sizeof(float)));
    CHECK(cudaStreamCreate(&stream_));

    delete[] trtModelStream;

    //cpu memory
    inputdata_ = new float[batchSize_ * input_size_];
    outputdata_ = new float[batchSize_ * output_size_];

    //结果保存
    for(int i = 0; i < batchSize_;i++){
        Results* results = new Results;
        results_map_.insert(std::make_pair(i,results));
    }
}

YoloV5::~YoloV5(){
    CHECK(cudaFree(buffers[0]));
    CHECK(cudaFree(buffers[1]));
    // Destroy the engine
    cudaStreamDestroy(stream_);
    context_->destroy();
    engine_->destroy();
    runtime_->destroy();

    delete[] inputdata_;
    delete[] outputdata_;

    for(int i = 0; i < batchSize_;i++){
        delete results_map_[i];
    }
}



void YoloV5::preprocess(std::vector<cv::Mat>& imgs){
    int hw = input_h_ * input_w_;
    int hw2 = 2 * input_h_ * input_w_;
    oriimgs.clear();

#pragma omp parallel for num_threads(16)
    for(int b = 0; b < batchSize_; b++){
        cv::Mat inputmat = imgs[b];
        oriimgs.push_back(inputmat.clone());
        cv::Mat pr_img = preprocess_img(inputmat);
        int i = 0;
        int tmp_loc = b * input_size_;

        for (int row = 0; row < input_h_; ++row) {
            uchar* uc_pixel = pr_img.data + row * pr_img.step;
            for (int col = 0; col < input_w_; ++col) {                                  
                inputdata_[tmp_loc + i] = (float)uc_pixel[2] / 255.0;
                inputdata_[tmp_loc + i + hw] = (float)uc_pixel[1] / 255.0;
                inputdata_[tmp_loc + i + hw2] = (float)uc_pixel[0] / 255.0;
                uc_pixel += 3;
                ++i;
            }
        }
    }
}

void YoloV5::infer(){
    CHECK(cudaMemcpyAsync(buffers[0], inputdata_, batchSize_ * input_size_ * sizeof(float), cudaMemcpyHostToDevice, stream_));
    context_->enqueue(batchSize_, buffers, stream_, nullptr);
    CHECK(cudaMemcpyAsync(outputdata_, buffers[1], batchSize_ * output_size_ * sizeof(float), cudaMemcpyDeviceToHost, stream_));
    cudaStreamSynchronize(stream_);
}

void YoloV5::post(){
    for(int i = 0; i < results_map_.size(); i++){
        results_map_[i]->results.clear(); 
    }
    std::vector<std::vector<Yolo::Detection>> batch_res(batchSize_);
    for (int b = 0; b < batchSize_; b++) {
        auto& res = batch_res[b];
        nms(res, &outputdata_[b * output_size_], 0.2, 0.5);
        reducerepeat(res);
        checkpeoplelocation(res);
        mergeperson(res);
    }
    
    for (int b = 0; b < batchSize_; b++) {
        auto& res = batch_res[b];
        cv::Mat inputmat = oriimgs[b];
        //Results tmpres;
        for (size_t j = 0; j < res.size(); j++) {
            cv::Rect r = get_rect(inputmat, res[j].bbox);
            int class_index = (int)res[j].class_id;	
            if(class_index == 20){
                class_index = 10;
            }

            MyResult result;
            result.num_class = class_index;
            result.location = {r.x, r.y, r.x + r.width, r.y+r.height};
            result.center_point = {(float)r.x + (float)r.width/2.0, (float)r.y+(float)r.height/2.0};
            result.bottom_center_point = {(float)r.x + (float)r.width/2.0, r.y + r.height};
            result.prob = res[j].conf;
            results_map_.at(b)->results.push_back(result);
        }
        results_map_.at(b)->index = b;
       
    }
}

std::map<int,Results*> YoloV5::GetResults(){
    return results_map_;
}
